import inspect


def grab_params(func):
    args = inspect.signature(func)
    args = list(args.parameters)
    args.remove('self')

    if 'kwargs' in args:
        args.remove('kwargs')

    return args


def get_mapping(lines, args):

    r = "        If \'default\', use the saved value within"
    r += " default params."

    new_lines = []
    for line in lines:
        if "If \'default\'" in line:
            new_lines.append(r)
        else:
            new_lines.append(line)
    lines = new_lines

    start_keys = ['    ' + a for a in args]
    start_lines = [[l for l in lines if l.startswith(key)][0]
                   for key in start_keys]
    start_inds = [lines.index(line) for line in start_lines]

    return_line = [l for l in lines if l.startswith('    Returns')]

    if len(return_line) > 0:
        return_ind = lines.index(return_line[0])
    else:
        return_ind = len(lines)

    start_inds.append(return_ind)

    mapping = {}
    for k in range(len(start_keys)):
        mapping[start_keys[k]] =\
            '\n'.join(lines[start_inds[k]:start_inds[k+1]])

    return mapping


def get_from_func(func):

    args = grab_params(func)
    doc = func.__doc__

    lines = doc.split('\n')
    mapping = get_mapping(lines, args)

    return doc, mapping


def get_new_docstring(o_func, r_func):

    o_doc, o_mapping = get_from_func(o_func)
    r_doc, r_mapping = get_from_func(r_func)

    for key in r_mapping:
        if key in o_mapping:
            r_doc = r_doc.replace(r_mapping[key], o_mapping[key])

    return r_doc


def get_name(obj):

    name = obj.__module__ + '.' + obj.__qualname__

    if '.<locals>.child' in name:
        name = obj.__parent_name__

    name = name.replace('.tree.tree', '.tree')
    name = name.replace('.tree.tree', '.tree')

    base_replace_list = ['logistic', 'gpc', 'gpr', 'classification',
                         'regression', 'coordinate_descent', 'sklearn',
                         'forest', 'classes', 'base', 'multilayer_perceptron',
                         'univariate_selection', 'minimum_difference', 'deskl',
                         'exponential', 'logarithmic', 'rrc', 'data',
                         'variance_threshold']

    for r in base_replace_list:
        name = name.replace('.' + r + '.', '.')

    splits = name.split('.')
    for split in splits:
        if split.startswith('_'):
            name = name.replace('.' + split + '.', '.')

    name = name.replace('BPt.extensions.Feat_Selectors.RFE_Wrapper',
                        'sklearn.feature_selection.RFE')

    return name


def get_scorer_name(obj):

    name = obj.__name__
    name = name.replace('_wrapper', '')
    name = 'sklearn.metrics.' + name

    return name



"""
_Validation.py
====================================
Main class extension file for defining validation and train test splits.
"""
from deslib.dcs.lca import LCA
import pandas as pd
import numpy as np
import os

from ..helpers.Data_Helpers import get_unique_combo_df, reverse_unique_combo_df


def _get_cv(self, cv_params, show=False, show_original=True, return_df=False):

    from ..helpers.CV import CV

    if isinstance(cv_params, CV):
        return cv_params

    train_only_subjects = cv_params.train_only_subjects
    train_only_loc = cv_params.train_only_loc

    if self.all_data is None:
        self._print('Calling Prepare_All_Data()',
                    'to change the default merge behavior',
                    'call it again!')
        self.Prepare_All_Data()

    if isinstance(train_only_subjects, str):
        if train_only_subjects == 'nan':
            train_only_subjects = self.Get_Nan_Subjects()

    subjects_to_use = None
    if train_only_loc is not None:
        subjects_to_use = train_only_loc

    elif train_only_subjects is not None:
        subjects_to_use = train_only_subjects

    train_only = self._get_subjects_to_use(subjects_to_use)
    train_only = np.array(list(train_only))

    groups = cv_params.groups
    stratify = cv_params.stratify

    df = None

    if groups is not None and stratify is not None:
        raise RuntimeError('Warning: BPt does not currently ',
                           'support groups and',
                           'stratify together!')

    if groups is not None:

        if isinstance(groups, str):
            groups = [groups]

        groups = [self.name_map[g] if g in self.name_map else g
                  for g in groups]
        groups = self._add_strat_u_name(groups)

        grp, l_e = get_unique_combo_df(self.all_data, groups)

        cv = CV(groups=grp, train_only=train_only)

        if show:
            # Note, groups has no return df, will set to None
            df = self._get_info_on(cv.groups, groups, 'groups', l_e,
                                   train_only)

    elif stratify is not None:

        if isinstance(stratify, str):
            stratify = [stratify]

        stratify = [self.name_map[s] if s in self.name_map else s
                    for s in stratify]

        targets = self._get_base_targets_names()

        to_stratify = []
        for s in stratify:

            # Check if target - basically want to only add strat name
            #  if not a target.
            if s in targets:

                # Add as is if target
                to_stratify.append(s)

            # Then it is a strat, so make sure to add name if not already added
            else:
                to_stratify.append(self._add_strat_u_name(s))

        # Get the unique combo of passed to_stratify if multiple
        strat, l_e = get_unique_combo_df(self.all_data, to_stratify)

        # Generate the CV object
        cv = CV(stratify=strat, train_only=train_only)

        # Optional if show
        if show:
            df = self._get_info_on(cv.stratify, to_stratify, 'stratify', l_e,
                                   train_only, show_original,
                                   return_df=return_df)

    # If only train only
    elif len(train_only) > 0:
        cv = CV(train_only=train_only)
        self._print(len(train_only), 'Train only subjects defined.')

    else:
        cv = CV()

    if return_df:
        return cv, df

    return cv


def Define_Validation_Strategy(self, cv=None, groups=None, stratify=None,
                               train_only_loc=None, train_only_subjects=None,
                               show=True, show_original=True, return_df=False):
    '''Define a validation strategy to be used during different train/test
    splits, in addition to model selection and model hyperparameter
    cross validation.
    See Notes for more info.

    Note, can also pass a cv params objects here.

    Parameters
    ----------
    cv :  :class:`CV` or None, optional
        If None, then skip, otherwise can pass a :class:`CV`
        object here, and the rest of the parameters will be skipped.

        ::

            default = None

    groups : str, list or None, optional
        In the case of str input, will assume the str to refer
        to a column key within the loaded strat data,
        and will assign it as a value to preserve groups by
        during any train/test or K-fold splits.
        If a list is passed, then each element should be a str,
        and they will be combined into all unique
        combinations of the elements of the list.

        ::

            default = None

    stratify : str, list or None, optional
        In the case of str input, will assume the str to refer
        to a column key within the loaded strat data, or a loaded target col.,
        and will assign it as a value to preserve
        distribution of groups by during any train/test or K-fold splits.
        If a list is passed, then each element should be a str,
        and they will be combined into all unique combinations of
        the elements of the list.

        Any target_cols passed must be categorical or binary, and cannot be
        float. Though you can consider loading in a float target as a strat,
        which will apply a specific k_bins, and then be valid here.

        In the case that you have a loaded strat val with the same name
        as your target, you can distinguish between the two by passing
        either the raw name, e.g., if they are both loaded as 'Sex',
        passing just 'Sex', will try to use the loaded target. If instead
        you want to use your loaded strat val with the same name - you have
        to pass 'Sex' + self.self.strat_u_name (by default this is '_Strat').

        ::

            default = None

    train_only_loc : str, Path or None, optional
        Location of a file to load in train_only subjects,
        where any subject loaded as train_only will be assigned to
        every training fold, and never to a testing fold.
        This file should be formatted as one subject per line.

        You can load from a loc and pass subjects, the subjects
        from each source will be merged.

        This parameter is compatible with groups / stratify.

        ::

            default = None

    train_only_subjects : set, array-like, 'nan', or None, optional
        An explicit list or array-like of train_only subjects, where
        any subject loaded as train_only will be assigned to every training
        fold, and never to a testing fold.

        You can also optionally specify 'nan' as input, which
        will add all subjects with any NaN data to train only.

        If you want to add both all the NaN subjects and custom
        subjects, call :func:`Get_Nan_Subjects` to get all NaN subjects,
        and then merge them yourself with any you want to pass.

        You can load from a loc and pass subjects, the subjects
        from each source will be merged.

        This parameter is compatible with groups / stratify.

        ::

            default = None

    show : bool, optional
        By default, if True, information about the defined validation
        strategy will be shown, including a dataframe if stratify is defined.

        ::

            default = True

    show_original : bool, optional
        By default when you define stratifying behavior, a dataframe will
        be displayed. This param controls if that dataframe shows original
        names, or if False, then it shows the internally used names.

        ::

            default = True

    return_df : bool, optional

        If set to true, then will return as dataframe version of
        the defined validation strategy. Note: this will return
        None in all cases execpt for when stratifying by a variable
        is requested!

        ::

            default = False

    Notes
    ----------
    Validation strategy choices are explained in more detail:

    - Random
        Just make validation splits randomly.

    - Group Preserving
        Make splits that ensure subjects that are
        part of specific group are all within the same fold
        e.g., split by family, so that people with the same family id
        are always a part of the same fold.

    - Stratifying
        Make splits such that the distribution of a given
        group is as equally split between two folds as possible,
        so simmilar to matched halves or
        e.g., in a binary or categorical predictive context,
        splits could be done to ensure roughly equal distribution
        of the dependent class.

    For now, it is possible to define only one overarching strategy
    (One could imagine combining group preserving splits
    while also trying to stratify for class,
    but the logistics become more complicated).
    Though, within one strategy it is certainly possible to
    provide multiple values
    e.g., for stratification you can stratify by target
    (the dependent variable to be predicted)
    as well as say sex, though with addition of unique value,
    the size of the smallest unique group decreases.
    '''

    if cv is not None:
        passed_cv = cv

    else:
        from .Params_Classes import CV
        passed_cv = CV(groups=groups, stratify=stratify,
                       train_only_loc=train_only_loc,
                       train_only_subjects=train_only_subjects)

    returned = self._get_cv(passed_cv, show=show,
                            show_original=show_original,
                            return_df=return_df)

    if return_df:
        self.cv, df = returned
    else:
        self.cv = returned
        df = None

    if return_df:
        return df


def Train_Test_Split(self, test_size=None,
                     test_subjects=None, cv='default',
                     random_state='default', test_loc='depreciated',
                     CV='depreciated'):
    '''Define the overarching train / test split, *highly reccomended*.

    Parameters
    ----------
    test_size : float, int or None, optional
        If float, should be between 0.0 and 1.0 and represent
        the proportion of the dataset to be included in the test split.
        If int, represents the absolute number (or target number) to
        include in the testing group.
        Keep as None if using test_subjects.

        ::

            default = None

    test_subjects : :ref:`Subjects`, optional
        Pass in a :ref:`Subjects` (see for more info) formatted input.
        This will define an explicit set of subjects to use as a test set.
        If anything but None is passed here, nothing should be passed to the
        test_size parameter.

        ::

            default = None

    cv : 'default' or :class:`CV`, optional
        If left as default 'default', use the class defined CV
        for the train test split, otherwise can pass custom behavior

        ::

            default = 'default'

    random_state : int None or 'default', optional
        If using test_size, then can optionally provide a random state, in
        order to be able to recreate an exact test set.

        If set to default, will use the value saved in self.random_state,
        (as set in :class:`BPt.BPt_ML` upon class init).

        ::

            default = 'default'

     test_loc : depreciated
        Pass a single str with the test loc to test_subjects instead.

        ::

            default = 'depreciated'

     CV : 'depreciated'
        Switching to passing cv parameter as cv instead of CV.
        For now if CV is passed it will still work as if it were
        passed as cv.

        ::

            default = 'depreciated'

    '''

    if test_loc != 'depreciated':
        print('Test loc is depreciated, please pass it as a single ',
              'string to test_subjects.')
        test_subjects = test_loc

    if CV != 'depreciated':
        print('Warning: Passing CV is depreciated. Please change to',
              'passing as cv instead.')

        # For now, let it still work
        cv = CV

    if cv == 'default':
        cv_obj = self.cv
    else:
        cv_obj = self._get_cv(cv)

    if test_size is None and test_subjects is None:
        test_size = .2

    if self.all_data is None:
        self._print('Calling Prepare_All_Data()',
                    'to change the default merge behavior',
                    'call it again!')
        self.Prepare_All_Data()

    if random_state == 'default':
        random_state = self.random_state

    if test_size is not None:

        _, subjects, train_only =\
            cv_obj.get_train_only(self.all_data.index)

        self._print('Performing split on', len(subjects), 'subjects', end='')

        if test_size == 0:
            self.train_subjects = subjects
            self.test_subjects = []
            self._print('Warning: Test size of 0 passed, all subjects set to',
                        'train.')

        else:

            if len(train_only) > 0:
                self._print(' with', len(train_only), 'considered train only!')
            else:
                self._print('!')

            self._print('random_state:', random_state)
            self._print('Test split size:', test_size)

            self.train_subjects, self.test_subjects = cv_obj.train_test_split(
                                    self.all_data.index, test_size,
                                    random_state)

        self._print()

    else:

        # Load passed subjects
        test_subjects = self._get_subjects_to_use(test_subjects)

        # Take only the overlap of the passed subjects with what is loaded
        test_subjects = [subject for subject in test_subjects
                         if subject in self.all_data.index]

        # Set remaining subjects to train subjects
        train_subjects = [subject for subject in self.all_data.index
                          if subject not in test_subjects]

        # Set the tr test subjects saved within class obj
        self.train_subjects = pd.Index(train_subjects,
                                       name=self.subject_id)
        self.test_subjects = pd.Index(test_subjects,
                                      name=self.subject_id)

    self._print('Performed train test split')
    self._print('Train size:', len(self.train_subjects))
    self._print('Test size: ', len(self.test_subjects))

    # Save train / test subjects if log_dr
    if self.log_dr is not None:

        train_loc = os.path.join(self.exp_log_dr, 'train_subjects.txt')
        test_loc = os.path.join(self.exp_log_dr, 'test_subjects.txt')

        for loc, subjects in zip([train_loc, test_loc], [self.train_subjects,
                                                         self.test_subjects]):
            with open(loc, 'w') as f:
                for subject in subjects:
                    f.write(str(subject) + '\n')


def _add_strat_u_name(self, in_vals):

    if in_vals is None:
        return None

    if isinstance(in_vals, str):
        if self.strat_u_name not in in_vals:
            new_vals = in_vals + self.strat_u_name
        else:
            new_vals = in_vals
    else:
        new_vals = []

        for val in in_vals:
            new_vals.append(self._add_strat_u_name(val))

    return new_vals


def _get_info_on(self, all_vals, col_names, v_type, l_e, train_only,
                 show_original=True, return_df=False):

    if v_type == 'groups':
        chunk = 'group preserving'
    elif v_type == 'stratify':
        chunk = 'stratifying behavior'
    else:
        chunk = 'error'

    if len(train_only) > 0:
        self._print(len(train_only), 'Train only subjects defined.')
        self._print('Those subjects are excluded from the below stats!')
        self._print()

        non_train_only = np.setdiff1d(all_vals.index, train_only,
                                      assume_unique=True)
        all_vals = all_vals.loc[non_train_only]

    unique_vals, counts = np.unique(all_vals, return_counts=True)

    self._print('CV defined with', chunk, 'over',
                len(unique_vals), 'unique values.')

    if v_type == 'stratify':

        if l_e is not None:
            col_split = reverse_unique_combo_df(unique_vals, l_e)

        else:
            col_names = [col_names]
            col_split = np.expand_dims(unique_vals, axis=-1)

        # Make a display df for each col name
        display_df = pd.DataFrame()
        for n in range(len(col_names)):
            name = col_names[n]

            if name in self.strat_encoders:
                encoder = self.strat_encoders[name]
                name = name.replace(self.strat_u_name, '')
            else:
                name = name.replace(self.strat_u_name, '')
                encoder = self.targets_encoders[name]

            if show_original:

                if isinstance(encoder, dict):
                    display_df[name] = [encoder[v] for v in col_split[:, n]]

                else:
                    try:
                        display_df[name] =\
                            encoder.inverse_transform(col_split[:, n])
                    except ValueError:
                        display_df[name] =\
                            np.squeeze(encoder.inverse_transform(
                                    col_split[:, n].reshape(-1, 1)))

            else:
                display_df[name] = col_split[:, n]

        display_df['Counts'] = counts
        self._display_df(display_df)

        # If return_df and strat, return display_df
        if return_df:
            return display_df

    # All other cases, return None
    return None



import tempfile
import random
import numpy as np
import os


def f_array(in_array, tp='float32'):
    return np.array(in_array).astype(tp)


def to_memmap(X):

    f = os.path.join(tempfile.gettempdir(), str(random.random()))
    x = np.memmap(f, dtype=X.dtype, shape=X.shape, mode='w+')
    x[:] = X

    return f, X.dtype, X.shape


def from_memmap(X):

    X_file, X_type, X_shape = X
    X = np.memmap(X_file, dtype=X_type, shape=X_shape, mode='c')
    return X


def is_ng(p):

    try:
        return 'nevergrad' in p.__module__
    except AttributeError:
        return False


def extract_values(value):

    if is_ng(value):

        # If a choice obj
        if hasattr(value, 'choices'):

            # Unpack choices
            choices = []
            for c in range(len(value.choices)):

                # Check for nested
                choice_value = extract_values(value.choices[c])
                if isinstance(choice_value, list):
                    choices += choice_value
                else:
                    choices.append(choice_value)

            return choices

        # if scalar type
        elif hasattr(value, 'integer'):

            # If cast to integer
            if value.integer:

                lower = value.bounds[0]
                if len(lower) == 1:
                    lower = int(lower[0])
                else:
                    lower = None

                upper = value.bounds[1]
                if len(upper) == 1:
                    upper = int(upper[0])
                else:
                    upper = None

                if lower is not None and upper is not None:
                    return list(range(lower, upper+1))

        elif hasattr(value, 'value'):
            return value.value

        # All other cases
        raise RuntimeError('Could not convert nevergrad',
                           value, 'to grid search parameter!')

    else:
        return value


def get_grid_params(params):

    # Set grid params
    grid_params = {}
    for p in params:
        grid_params[p] = extract_values(params[p])

    return grid_params



"""
_Plotting.py
====================================
Main class extension file for the some plotting functionality.
"""
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import math
import os
from IPython.display import display, HTML
from matplotlib.animation import FuncAnimation
from ..helpers.Data_File import load_data_file_proxies

from ..helpers.Data_Helpers import get_original_cat_names


def proc_title_length(title, br='-\n'):

    LIM = 55

    new_title = ''
    for i in range(LIM, len(title)+(LIM-1), LIM):

        if i >= len(title):
            new_title += title[i-LIM:]
        else:
            new_title += title[i-LIM:i] + br

    return new_title


def _plot(self, save_name, show=True):

    if show:
        if self.log_dr is not None:

            save_name = save_name.replace(' ', '_')
            save_name = save_name.replace('/', '')

            save_spot = os.path.join(self.exp_log_dr,
                                     save_name.replace(' ', '_') + '.png')
            plt.savefig(save_spot, dpi=self.dpi, bbox_inches='tight')

        if self.notebook:
            plt.show()
        else:
            plt.close()


def _proc_subjects(self, data, subjects):

    if subjects is None:
        return data

    elif subjects == 'train':

        try:
            return data.loc[self.train_subjects].copy()
        except KeyError:
            raise RuntimeError('No train subjects defined!')

    elif subjects == 'test':

        try:
            return data.loc[self.test_subjects].copy()
        except KeyError:
            raise RuntimeError('No test subjects defined!')

    elif subjects == 'both':

        try:
            return (data.loc[self.train_subjects].copy(),
                    data.loc[self.test_subjects].copy())
        except KeyError:
            raise RuntimeError('No train/test subjects defined!')

    else:

        try:
            return data.loc[subjects].copy()
        except KeyError:
            raise RuntimeError('Invalid subjects passed!')


def _get_plot_df(self, source_df, subjects, show_only_overlap):

    if len(source_df) == 0:
        df = self.all_data.copy()
    else:
        df = self._set_overlap(source_df, show_only_overlap).copy()

    return self._proc_subjects(df, subjects)


def _get_col(data, i):

    col = data[list(data)[i]]
    non_nan_col = col[~pd.isnull(col)]
    n_nan_subjects = len(col) - len(non_nan_col)

    return non_nan_col, n_nan_subjects


def _plot_seaborn_dist(data, plot_type, label=None):

    if plot_type == 'kde':
        sns.kdeplot(data, label=label)
    elif plot_type == 'bar':
        sns.boxplot(data, label=label)
    else:
        try:
            sns.distplot(data, label=label)
        except RuntimeError:
            sns.distplot(data, kde=False, label=label)


def Show_Data_Dist(self, data_subset='SHOW_ALL',
                   num_feats=20, feats='random',
                   reduce_func=None,
                   frame_interval=500,
                   plot_type='hist', show_only_overlap=True,
                   subjects=None, save=True, dpi='default',
                   save_name='data distribution', random_state='default',
                   return_anim=False):

    '''This method displays some summary statistics about
    the loaded targets, as well as plots the distibution if possible.

    Note: to display loaded data files, pass a fun to reduce_func, otherwise
    they will not be displayed.

    Parameters
    ----------
    data_subset : 'SHOW_ALL' or array-like, optional
        'SHOW_ALL' is reserved for showing
        the distributions of loaded data.
        You may also pass a list/array-like to specify specific
        a custom source of features to show.

        If self.all_data is already prepared, this data subset can also include
        any float type features loaded as covar or target.

        ::

            default = 'SHOW_ALL'

    num_feats: int, optional
        The number of features' distributions in which to view.
        Note: If too many are selected it may take a long time to render
        and/or consume a lot of memory!

        ::

            default = 20

    feats : {'random', 'skew'}, optional
        The features in which to display, if 'random' then
        will select `num_feats` random features to display.
        If 'skew', will show the top `num_feats` features by
        absolute skew.

        If 'skew' and subjects == 'both', will
        compute the top skewed features based on
        the training set.

        ::

            default = 'random'

    reduce_func : python function or list of, optional
        If a function is passed here, then data files will be loaded
        and reduced to 1 number according to the passed function.
        For example, the default function is just to take the
        mean of each loaded file, and to compute outlier detection
        on the mean.

        To not display data files, if any, then just keep
        reduce func as None

        ::

            default = None

    frame_interval: int, optional
        The number of milliseconds between each frame.

        ::

            default = 500

    plot_type : {'bar', 'hist', 'kde'}
        The type of base seaborn plot to generate for each datapoint.
        Either 'bar' for barplot, or 'hist' for seaborns dist plot, or
        'kde' for just a kernel density estimate plot.

        ::

            default = 'hist'

    show_only_overlap : bool, optional
        If True, then displays only the distributions for valid overlapping
        subjects across data, covars, ect... otherwise, if False,
        shows the current loaded distribution as is.

        If subjects is set (anything but None), this param will be ignored.

        ::

            default = True

    subjects : None, 'train', 'test', 'both' or array-like, optional
        If not None, then plot only the subjects loaded as train_subjects,
        or as test subjects, of you can pass a custom list or array-like of
        subjects.

        If 'both', then will plot the train and test distributions seperately.
        Note: This only works for plot_type == 'hist' or 'kde'.
        Also take into account, specifying 'both' will show some
        different information, then the default settings.

        ::

            default = None

    save : bool, optional
        If the animation should be saved as a gif, True or False.

        ::

            default = True

    dpi : int, 'default', optional
        The dpi in which to save the distribution gif.
        If 'default' use the class default value.

        ::

            default = 'default'

    save_name : str, optional
        The name in which the gif should be saved under.

        ::

            default = 'data distribution'

    random_state : 'default', int or None
        The random state in which to choose random features from.
        If 'default' use the class define value, otherwise set to the value
        passed. None for random.

        ::

            default = 'default'

    return_anim : bool, optional
        If True, return just the animation

        ::

            default = False

    '''

    # If data in low memory work for all data instead
    if len(self.data) == 0:
        valid_data = self.all_data.copy()[self.Data_Scopes.data_keys]
    else:
        valid_data = self.data.copy()
        if subjects is None:
            valid_data = self._set_overlap(valid_data, show_only_overlap)

    # If not all passed data subset are in data, try all data instead
    if data_subset != 'SHOW_ALL':
        if all([feat in valid_data for feat in data_subset]):
            valid_data = valid_data[data_subset]
        else:
            valid_data = self.all_data.copy()[data_subset]

    # If loading data files also
    if reduce_func is None:
        valid_data = valid_data.drop(self.data_file_keys, axis=1)
    else:
        data_file_data =\
            load_data_file_proxies(valid_data, [reduce_func],
                                   self.data_file_keys,
                                   self.file_mapping,
                                   n_jobs=self.n_jobs)[0]

        valid_data = valid_data.drop(self.data_file_keys, axis=1)
        valid_data = pd.merge(valid_data, data_file_data, on=self.subject_id)

    if random_state == 'default':
        random_state = self.random_state

    self._print('Plotting data distribution.')

    if subjects == 'both':
        if plot_type == 'bar':
            raise RuntimeWarning('Switching plot type to dist due to subjects',
                                 ' == "both"')
            plot_type = 'dist'

        data, test_data = self._proc_subjects(valid_data, subjects)

        self._print('Viewing train data with shape:', data.shape)
        self._print('Viewing test data with shape:', test_data.shape)

    else:

        data = self._proc_subjects(valid_data, subjects)

        self._print('Viewing data with shape:', data.shape)
        self._print()

        self._print('Loaded data top columns by skew:')
        self._print(valid_data.skew().sort_values())
        self._print()

    fig = plt.figure()

    def update(i):
        fig.clear()
        title = ''

        if subjects == 'both':
            non_nan_col_tr, n_tr = _get_col(data, i)
            non_nan_col_test, n_test = _get_col(test_data, i)

            if n_tr > 0 or n_test > 0:
                title = 'Train - NaN subjects not shown: ' + str(n_tr)
                title += '\nTest - NaN subjects not shown: ' + str(n_test)

            _plot_seaborn_dist(non_nan_col_tr, plot_type, label='train')
            _plot_seaborn_dist(non_nan_col_test, plot_type, label='test')

            plt.legend()

        else:
            non_nan_col, n = _get_col(data, i)

            if n > 0:
                title = 'NaN subjects not shown: ' + str(n)

            _plot_seaborn_dist(non_nan_col, plot_type)

        plt.title(title, fontdict={'fontsize': 'medium'})

    if 'skew':
        most_skewed = data.skew().abs().sort_values()[-num_feats:].index
        frames = [list(data).index(m) for m in most_skewed][::-1]
    else:
        np.random.seed(random_state)
        frames = np.random.randint(0, data.shape[1], size=num_feats)

    anim = FuncAnimation(fig, update, frames=frames, interval=500)
    if return_anim:
        return anim

    try:
        html = HTML(anim.to_html5_video())
    except RuntimeError:
        print('To see a gif of the data distribution, make sure you '
              'have ffmpeg installed!')
        return None

    if self.log_dr is not None:

        save_name = os.path.join(self.exp_log_dr,
                                 save_name.replace(' ', '_') + '.gif')

        try:
            anim.save(save_name, dpi=self.dpi, writer='imagemagick')
        except BrokenPipeError:
            print('Warning: could not save gif, please make sure you have',
                  'imagemagick installed')
        plt.close()

    if self.notebook:
        plt.close()
        return html

    return None


def _input_targets(self, targets):

    if targets == 'SHOW_ALL':
        targets = self._get_base_targets_names()

    if not isinstance(targets, list):
        targets = [targets]

    targets = [self._get_targets_key(t) for t in targets]

    return targets


def _input_covars(self, covars):

    if covars == 'SHOW_ALL':
        covars = list(self.covars_encoders)

    if not isinstance(covars, list):
        covars = [covars]

    return covars


def _input_strat(self, strat):

    if strat == 'SHOW_ALL':
        strat = list(self.strat_encoders)

    if not isinstance(strat, list):
        strat = [strat]

    strat = [self._add_strat_u_name(s) for s in strat]

    return strat


def Show_Targets_Dist(self, targets='SHOW_ALL', cat_show_original_name=True,
                      show_only_overlap=True, subjects=None, show=True,
                      cat_type='Counts', return_display_dfs=False):
    '''This method displays some summary statistics about
    the loaded targets, as well as plots the distibution if possible.

    Parameters
    ----------
    targets : str, int or list, optional
        The single (str) or multiple targets (list),
        in which to display the distributions of. The str input
        'SHOW_ALL' is reserved, and set to default, for showing
        the distributions of loaded targets.

        You can also pass the int index of the loaded target to show!

        (default = 'SHOW_ALL')

    cat_show_original_name : bool, optional
        If True, then when showing a categorical distribution (or binary)
        make the distr plot using the original names. Otherwise,
        use the internally used names.

        (default = True)

    show_only_overlap : bool, optional
        If True, then displays only the distributions for valid overlapping
        subjects across data, covars, ect... otherwise, if False,
        shows the current loaded distribution as is.

        (default = True)

    subjects : None, 'train', 'test' or array-like, optional
        If None, plot all subjects.
        If not None, then plot only the subjects loaded as train_subjects,
        or as test subjects, or you can pass a custom list or array-like of
        subjects.

        (default = None)

    show : bool, optional
        If True, then plt.show(), the matplotlib command will be called,
        and the figure displayed. On the other hand, if set to False,
        then the user can customize the plot as they desire.
        You can think of plt.show() as clearing all of the loaded settings,
        so in order to make changes, you can't call this until you are done.

        (default = True)

    cat_type : {'Counts', 'Frequency'}, optional

        If plotting a categorical variable (binary or categorical),
        plot the X axis as either by raw count or frequency.

        (default = 'Counts')

    return_display_dfs : bool, optional

        Optionally return the display df as a pandas df

        (default = False)
    '''

    targets_df = self._get_plot_df(self.targets, subjects, show_only_overlap)
    targets = self._input_targets(targets)

    display_dfs = []
    for target in targets:
        display_df = self._show_single_dist(target, targets_df,
                                            self.targets_encoders,
                                            cat_show_original_name, show,
                                            source='target',
                                            cat_type=cat_type)
        display_dfs.append(display_df)
    self._print()

    if return_display_dfs:
        return display_dfs


def Show_Covars_Dist(self, covars='SHOW_ALL', cat_show_original_name=True,
                     show_only_overlap=True, subjects=None, show=True,
                     cat_type='Counts', return_display_dfs=False):
    '''Plot a single or multiple covar distributions, along with
    outputting useful summary statistics.

    Parameters
    ----------
    covars : str or list, optional
        The single covar (str) or multiple covars (list),
        in which to display the distributions of. The str input
        'SHOW_ALL' is reserved, and set to default, for showing
        the distributions of all loaded covars.

        (default = 'SHOW_ALL')

    cat_show_original_name : bool, optional
        If True, then when showing a categorical distribution (or binary)
        make the distr plot using the original names. Otherwise,
        use the internally used names.

        (default = True)

    show_only_overlap : bool, optional
        If True, then displays only the distributions for valid overlapping
        subjects across data, covars, ect... otherwise, shows the current
        loaded distribution as is.

        (default = True)

    subjects : None, 'train', 'test' or array-like, optional
        If not None, then plot only the subjects loaded as train_subjects,
        or as test subjects, of you can pass a custom list or array-like of
        subjects.

        (default = None)

    show : bool, optional
        If True, then plt.show(), the matplotlib command will be called,
        and the figure displayed. On the other hand, if set to False,
        then the user can customize the plot as they desire.
        You can think of plt.show() as clearing all of the loaded settings,
        so in order to make changes, you can't call this until you are done.

        (default = True)

    cat_type : {'Counts', 'Frequency'}, optional

        If plotting a categorical variable (binary or categorical),
        plot the X axis as either by raw count or frequency.

        (default = 'Counts')

    return_display_dfs : bool, optional

        Optionally return the display df as a pandas df

        (default = False)
    '''

    covars_df = self._get_plot_df(self.covars, subjects, show_only_overlap)
    covars = self._input_covars(covars)

    display_dfs = []
    for covar in covars:
        display_df = self._show_single_dist(covar, covars_df,
                                            self.covars_encoders,
                                            cat_show_original_name, show,
                                            source='covar',
                                            cat_type=cat_type)
        display_dfs.append(display_df)
        self._print()

    if return_display_dfs:
        return display_dfs


def Show_Strat_Dist(self, strat='SHOW_ALL', cat_show_original_name=True,
                    show_only_overlap=True, subjects=None, show=True,
                    cat_type='Counts', return_display_dfs=False):
    '''Plot a single or multiple strat distributions, along with
    outputting useful summary statistics.

    Parameters
    ----------
    strat : str or list, optional
        The single strat (str) or multiple strats (list),
        in which to display the distributions of. The str input
        'SHOW_ALL' is reserved, and set to default, for showing
        the distributions of all loaded strat cols.

        (default = 'SHOW_ALL')

    cat_show_original_name : bool, optional
        If True, then make the distr. plot using the original names.
        Otherwise, use the internally used names.

        (default = True)

    show_only_overlap : bool, optional
        If True, then displays only the distributions for valid overlapping
        subjects across data, covars, ect... otherwise, shows the current
        loaded distribution as is.

        (default = True)

    subjects : None, 'train', 'test' or array-like, optional
        If not None, then plot only the subjects loaded as train_subjects,
        or as test subjects, of you can pass a custom list or array-like of
        subjects.

        (default = None)

    show : bool, optional
        If True, then plt.show(), the matplotlib command will be called,
        and the figure displayed. On the other hand, if set to False,
        then the user can customize the plot as they desire.
        You can think of plt.show() as clearing all of the loaded settings,
        so in order to make changes, you can't call this until you are done.

        (default = True)

    cat_type : {'Counts', 'Frequency'}, optional

        If plotting a categorical variable (binary or categorical),
        plot the X axis as either by raw count or frequency.

        (default = 'Counts')

    return_display_dfs : bool, optional

        Optionally return the display df as a pandas df

        (default = False)
    '''

    strat_df = self._get_plot_df(self.strat, subjects, show_only_overlap)
    strat = self._input_strat(strat)

    display_dfs = []
    for s in strat:
        display_df = self._show_single_dist(s, strat_df, self.strat_encoders,
                                            cat_show_original_name, show,
                                            source='strat',
                                            cat_type=cat_type)
        display_dfs.append(display_df)
        self._print()

    if return_display_dfs:
        return display_dfs


def _get_single_df(self, name, df, all_encoders):

    try:
        encoder = all_encoders[name]
    except KeyError:
        raise KeyError('No col named', name, 'found!')

    # list = multilabel
    # 3-tuple = one-hot or dummy
    # OrdinalEncoder = ordinal categorical
    # None = regression
    # dict = Binary conv'ed from regression

    # Regression
    if encoder is None:
        single_df = df[[name]].copy()

    # Multilabel
    elif isinstance(encoder, list):
        single_df = df[encoder].copy()

    # Binary/ordinal
    else:
        single_df = df[[name]].copy()

    return single_df, encoder


def _show_single_dist(self, name, df, all_encoders, cat_show_original_name,
                      show=True, source='target', cat_type='Counts', alpha=1,
                      color=None, label=None):

    # If subjects passed as both
    if len(df) == 2:

        # Temp mute verbose + log_file
        verbose, log_file, notebook =\
            self.verbose, self.log_file, self.notebook
        self.verbose, self.log_file, self.notebook = False, None, False

        train_df = self._show_single_dist(name, df[0],
                                          all_encoders,
                                          cat_show_original_name,
                                          show=False,
                                          source='target',
                                          cat_type=cat_type,
                                          label='Train')
        test_df = self._show_single_dist(name, df[1],
                                         all_encoders,
                                         cat_show_original_name,
                                         show=False,
                                         source='target',
                                         cat_type=cat_type,
                                         label='Test')

        # Add legend!
        plt.legend()

        self.verbose, self.log_file, self.notebook =\
            verbose, log_file, notebook

        # If binary or cat
        if train_df.index.name == 'Internal_Name':

            train_df['Split'] = 'Train'
            test_df['Split'] = 'Test'
            merged_df = pd.concat([train_df, test_df])

            # Clear existing
            plt.clf()

            y = 'Internal_Name'
            if 'Original_Name' in list(merged_df):
                y = 'Original_Name'

            sns.barplot(x=cat_type, y=y, hue='Split',
                        data=merged_df.reset_index(), orient='h')
        else:
            name = list(train_df)[0]
            merged_df = train_df.rename({name: 'Train'}, axis=1)
            merged_df['Test'] = test_df[name]
            merged_df.index.name = name

        self._display_df(merged_df)

        title = name + ' ' + source + ' distribution'
        plt_title = title.replace(' ' + source, '')
        plt_title = plt_title.replace(self.strat_u_name, '')
        plt_title = proc_title_length(plt_title)
        plt.title(plt_title)
        self._plot(title, show)

        return merged_df

    else:

        single_df, encoder =\
            self._get_single_df(name, df, all_encoders)

        return self._show_dist(single_df, plot_key=name,
                               cat_show_original_name=cat_show_original_name,
                               encoder=encoder, original_key=name,
                               show=show,
                               source=source, cat_type=cat_type,
                               alpha=alpha, color=color, label=label)


def _get_cat_display_df(self, df, encoder, name, cat_show_original_name):

    # Multilabel
    if isinstance(encoder, list):
        cat_show_original_name = False
        sums = df.sum()

    # Binary/ordinal
    else:
        unique, counts = np.unique(df, return_counts=True)
        sums = pd.Series(counts, index=unique)

        # This portion is to ensure that
        # every index with a count of 0 is filled in with 0's
        # This works as all binary / ordinal's internal index
        # go 0 to max_sum - 1, even those with saved dict encoding
        if isinstance(encoder, dict):
            max_sum = len(encoder)
        else:
            try:
                max_sum = encoder.n_bins
            except AttributeError:
                max_sum = len(encoder.classes_)

        for i in range(max_sum):
            if i not in sums.index:
                sums = sums.append(pd.Series([0], index=[i]))
        sums = sums.sort_index()

    display_df = pd.DataFrame(sums, columns=['Counts'])
    display_df.index.name = 'Internal_Name'
    display_df['Frequency'] = sums / len(df)

    original_names = None
    if cat_show_original_name:

        original_names = get_original_cat_names(sums.index,
                                                encoder,
                                                name)

        display_df['Original_Name'] = original_names
        display_df = display_df[['Original_Name', 'Counts', 'Frequency']]

    else:
        display_df = display_df[['Counts', 'Frequency']]

    return display_df, sums, original_names


def _show_dist(
 self, data, plot_key, cat_show_original_name, encoder=None, original_key=None,
 dropped_name=None, show=True, source='target', cat_type='Counts', alpha=1,
 color=None, label=None):

    # Ensure works with NaN data loaded
    no_nan_subjects = data[~data.isna().any(axis=1)].index
    nan_subjects = data[data.isna().any(axis=1)].index
    no_nan_data = data.loc[no_nan_subjects]

    self._print('--', plot_key, '--')

    # Regression
    if encoder is None:

        display_df = no_nan_data.describe()

        median = no_nan_data.median()
        median.name = 'median'
        display_df = display_df.append(median)

        skew = no_nan_data.skew()
        skew.name = 'skew'
        display_df = display_df.append(skew)

        kurtosis = no_nan_data.kurtosis()
        kurtosis.name = 'kurtosis'
        display_df = display_df.append(kurtosis)

        display_df = display_df.round(3)

        self._display_df(display_df)

        try:
            sns.distplot(no_nan_data, color=color, label=label)
        except RuntimeError:
            sns.distplot(no_nan_data, kde=False, color=color, label=label)

    # Binary/ordinal
    else:

        display_df, sums, original_names =\
            self._get_cat_display_df(no_nan_data, encoder, original_key,
                                     cat_show_original_name)
        if original_names is None:
            cat_show_original_name = False

        self._display_df(display_df)

        display_names = sums.index
        if cat_show_original_name:
            display_names = pd.Index(original_names)
            display_names.name = 'Original_Name'

        sns.barplot(x=display_df[cat_type], y=display_names, orient='h',
                    alpha=alpha, color=color, label=label)
        plt.xlabel(cat_type)

    # If any NaN
    if len(nan_subjects) > 0:
        self._print('Note:', len(nan_subjects), 'subject(s) with NaN',
                    'not included/shown!')

    title = plot_key + ' ' + source + ' distribution'
    plt_title = title.replace(' ' + source, '').replace(self.strat_u_name, '')
    plt_title = proc_title_length(plt_title)
    plt.title(plt_title)

    self._plot(title, show)

    return display_df


def _display_df(self, display_df):

    if self.notebook:
        display(display_df)

    self._print(display_df, dont_print=self.notebook)
    self._print(dont_print=self.notebook)


def _get_top_global(self, df, top_n, get_abs):

    if get_abs:
        imps = np.mean(np.abs(df))
    else:
        imps = np.mean(df)

    imps.sort_values(ascending=False, inplace=True)

    to_get = list(range(top_n))
    if not get_abs:
        to_get += [-i for i in range(1, top_n+1)][::-1]

    top = imps[to_get]
    feats = imps[to_get].index

    top_df = df[feats]

    if get_abs:
        top_df = top_df.abs()

    wide = top_df.melt()
    wide['mean'] = wide['variable'].copy()
    wide['mean'] = wide['mean'].replace(top.to_dict())

    return wide


def Plot_Global_Feat_Importances(
 self, feat_importances='most recent', top_n=10, show_abs=False,
 multiclass=False, ci=95, palette='default', figsize=(10, 10),
 title='default', titles='default', xlabel='default', n_cols=1, ax=None,
 show=True):
    '''Plots any global feature importance, e.g. base or shap, values per
    feature not per prediction.

    Parameters
    ----------
    feat_importances : 'most recent' or Feat_Importances object
        Input should be either a Feat_Importances object as output from a
        call to Evaluate, or Test, or if left as default 'most recent',
        the passed params will be used to plot any valid calculated feature
        importances from the last call to Evaluate or Test.

        Note, if there exist multiple valid feature importances in the
        last call, passing custom ax will most likely break things.

        (default = 'most recent')

    top_n : int, optional
        The number of top features to display. In the case where
        show_abs is set to True, or the feature importance being plotted
        is only positive, then top_n features will be shown. On the other
        hand, when show_abs is set to False and the feature importances being
        plotted contain negative numbers, then the top_n highest and top_n
        lowest features will be shown.

        (default = 10)

    show_abs : bool, optional
        In the case where the underlying feature importances contain
        negative numbers, you can either plot the top_n by absolute value,
        with show_abs set to True, or plot the top_n highest and lowest
        with show_abs set to False.

        (default = False)

    multiclass : bool, optional
        If multiclass is set to True, and the underlying feature importances
        were derived from a categorical problem type, then
        a seperate feature importance plot will be made for each class.
        Alternatively, if multiclass is set to False, then feature importances
        will be averaged over all classes.

        (default = False)

    ci : float, 'sd' or None, optional
        Size of confidence intervals to draw around estimated values.
        If 'sd', skip bootstrapping and draw the standard deviation
        of the feat importances. If None, no bootstrapping will be performed,
        and error bars will not be drawn.

        (default = 95)

    palette : Seaborn palette name, optional
        Color scheme to use. Search seaborn palettes for more information.
        Default for absolute is 'Reds', and default for both pos
        and neg is 'coolwarm'.

        (default = 'default')

    title : str, optional
        The title used during plotting, and also used
        to save a version of the figure
        (with spaces in title replaced by _, and as a png).

        When multiclass is True, this is the full figure title.

        (default = 'default')

    titles: list, optional
        This parameter is only used when multiclass is True.
        titles should be a list with the name for each classes plot.
        If left as default, it will just be named the original loaded name
        for that class.

        (default = 'default')

    xlabel : str, optional
        The xlabel, descriping the measure of feature importance.
        If left as 'default' it will change depend on what feature importance
        is being plotted.

        (default = 'default')

    n_cols: int, optional
        If multiclass, then the number of class plots to
        plot on each row.

        (default = 1)

    ax: matplotlib axis, or axes, optional
        A custom ax to plot to for an individual plot, or if using
        multiclass, then a list of axes can be passed here.

        (default = None)

    show : bool, optional
        If True, then plt.show(), the matplotlib command will be called,
        and the figure displayed. On the other hand, if set to False,
        then the user can customize the plot as they desire.
        You can think of plt.show() as clearing all of the loaded settings,
        so in order to make changes, you can't call this until you are done.

        (default = True)
    '''

    if feat_importances == 'most recent':
        for fis in self.evaluator.feat_importances:
            if 'global' in fis.scopes:
                self.Plot_Global_Feat_Importances(
                 fis, top_n, show_abs, multiclass, ci, palette, figsize, title,
                 titles, xlabel, n_cols, ax, show)

        return

    # Initial check to make sure valid feat importances passed
    if feat_importances.global_df is None:
        raise AttributeError('You must pass a feature importances object with',
                             'global_df, try plot local feat importance.')

    # If multiclass
    if isinstance(feat_importances.global_df, list):
        if multiclass:

            dfs = feat_importances.global_df

            return self._plot_multiclass_global_feat_importances(
                        dfs, feat_importances, top_n=top_n, show_abs=show_abs,
                        ci=ci, palette=palette, figsize=figsize, title=title,
                        titles=titles, xlabel=xlabel, n_cols=n_cols, ax=ax,
                        show=show)

        else:
            df = pd.concat(feat_importances.global_df)
    else:
        df = feat_importances.global_df

    # Non-multi-class case
    ax = self._plot_global_feat_importances(df, feat_importances,
                                            top_n=top_n, show_abs=show_abs,
                                            ci=ci, palette=palette,
                                            title=title, xlabel=xlabel,
                                            ax=ax, show=show)
    return ax


def _plot_multiclass_global_feat_importances(self, dfs, feat_importances,
                                             top_n=10, show_abs=False,
                                             ci=95, palette='default',
                                             figsize=(10, 10), title='default',
                                             titles='default',
                                             xlabel='default', n_cols=2,
                                             ax=None, show=True):

    # Grab the right classes
    target = feat_importances.target
    target_key = self._get_targets_key(target)
    classes = self.targets_encoders[target_key].classes_

    # Set default titles
    if titles == 'default':
        titles = classes

    if title == 'default':
        name = feat_importances.get_name()
        title = name + ' Top Features by Class'

    # If an ax is passed, then use that instead of making subplots
    n_rows = math.ceil(len(classes) / n_cols)
    if ax is not None:
        axes = ax
    else:
        fig, axes = plt.subplots(n_rows, n_cols, figsize=figsize)

    # In the case where only 1 row or 1 col.
    single = False
    if n_rows == 1 or n_cols == 1:
        single = True

    for i in range(len(dfs)):

        row = i // n_cols
        col = i % n_cols
        rc = (row, col)

        if single:
            rc = i

        # Create the subplot
        axes[rc] = self._plot_global_feat_importances(
            dfs[i], feat_importances, top_n=top_n,
            show_abs=show_abs, ci=ci, palette=palette,
            title=titles[i], xlabel='default', ax=axes[rc])

    # Set any remaining subplots to nothing
    for i in range(len(dfs), n_rows * n_cols):

        row = i // n_cols
        col = i % n_cols
        rc = (row, col)

        if single:
            rc = i

        axes[rc].set_axis_off()

    # Set the title and layout if no custom ax passed
    if ax is None:
        fig.tight_layout()
        fig.subplots_adjust(top=.9, hspace=.25)
        fig.suptitle(title, fontsize="x-large")
        self._plot(title, show)

    else:
        return axes


def _plot_global_feat_importances(self, df, feat_importances, top_n=10,
                                  show_abs=False, ci=95,
                                  palette='default', title='default',
                                  xlabel='default', ax=None, show=True):

    # Distinguish between feat importances that have been set to abs
    min_val = np.min(np.min(df))
    add_abs_sign = True

    if min_val >= 0:
        show_abs = True
    else:
        add_abs_sign = False

    # Get df, with only top_n, in wide form for plotting
    top_df = self._get_top_global(df, top_n, show_abs)

    # Abs / all pos values
    if show_abs:
        if palette == 'default':
            palette = 'Reds'

        if title == 'default':
            name = feat_importances.get_name()
            title = name + ' Top ' + str(top_n) + ' Features'

        if xlabel == 'default':
            xlabel = feat_importances.get_global_label()

        sns.barplot(x='value', y='variable', hue='mean',
                    data=top_df, orient='h', ci=ci,
                    palette=palette, errwidth=2, ax=ax, dodge=False)

    # Pos + neg values
    else:
        if palette == 'default':
            palette = 'coolwarm'

        if title == 'default':
            name = feat_importances.get_name()
            title = name + ' Top ' + str(top_n) + ' +/- Features'

        if xlabel == 'default':
            xlabel = feat_importances.get_global_label()

            if add_abs_sign:
                xlabel = '|' + xlabel + '|'

        sns.barplot(x='value', y='variable', hue='mean',
                    data=top_df, orient='h', ci=ci,
                    palette=palette, errwidth=2, ax=ax,
                    dodge=False)

        # Add an extra line down the center
        if ax is None:
            current_ax = plt.gca()
            current_ax.axvline(0, color='k', lw=.2)
            sns.despine(ax=current_ax, top=True, left=True, bottom=True,
                        right=True)
        else:
            ax.axvline(0, color='k', lw=.2)
            sns.despine(ax=ax, top=True, left=True, bottom=True, right=True)

    # Process rest fo params based on if for axis or global
    if ax is None:
        plt.title(title)
        plt.xlabel(xlabel)
        plt.ylabel('')
        plt.legend().remove()
        self._plot(title, show)

    else:
        ax.set_title(title)
        ax.set_xlabel(xlabel)
        ax.set_ylabel('')
        ax.get_legend().remove()
        return ax


def Plot_Local_Feat_Importances(self, feat_importances='most recent', top_n=10,
                                title='default', titles='default',
                                xlabel='default', one_class=None, show=True):
    '''Plots any local feature importance, e.g. shap, values per
    per prediction.

    Parameters
    ----------
    feat_importances : 'most recent' or Feat_Importances object
        Input should be either a Feat_Importances object as output from a
        call to Evaluate, or Test, or if left as default 'most recent',
        the passed params will be used to plot any valid calculated feature
        importances from the last call to Evaluate or Test.

        (default = 'most recent')

    top_n : int, optional
        The number of top features to display. In the case where
        show_abs is set to True, or the feature importance being plotted
        is only positive, then top_n features will be shown. On the other
        hand, when show_abs is set to False and the feature importances being
        plotted contain negative numbers, then the top_n highest and top_n
        lowest features will be shown.

        (default = 10)

    title : str, optional
        The title used during plotting, and also used
        to save a version of the figure
        (with spaces in title replaced by _, and as a png).

        With a multiclass / categorical problem type, this
        is only used if one_class is set. Otherwise, titles are used.

        (default = 'default')

    titles: list, optional
        This parameter is only used with a multiclass problem type.
        titles should be a list with the name for each class to plot.
        If left as default, it will use originally loaded class names.
        for that class.

        (default = 'default')

    xlabel : str, optional
        The xlabel, descriping the measure of feature importance.
        If left as 'default' it will change depend on what feature importance
        is being plotted.

        (default = 'default')

    one_class : int or None, optional
        If an underlying multiclass or categorical type, optionally
        provide an int here, corresponding to the single class to
        plot. If left as None, with make plots for all classes.

        (default = None)

    show : bool, optional
        If True, then plt.show(), the matplotlib command will be called,
        and the figure displayed. On the other hand, if set to False,
        then the user can customize the plot as they desire.
        You can think of plt.show() as clearing all of the loaded settings,
        so in order to make changes, you can't call this until you are done.

        (default = True)
    '''

    if feat_importances == 'most recent':
        for fis in self.evaluator.feat_importances:
            if 'local' in fis.scopes:
                self.Plot_Local_Feat_Importances(
                 fis, top_n, title, titles, xlabel, one_class, show)

        return

    if feat_importances.local_df is None:
        raise AttributeError('You must pass a feature importances object with',
                             'local_df, try plot global feat importance.')
        return

    df = feat_importances.local_df

    # If multiclass
    if isinstance(df, list):

        target_key = self._get_targets_key(feat_importances.target)
        classes = self.targets_encoders[target_key].classes_

        if titles == 'default':
            base = ' Shap ' + str(top_n) + ' Features'
            titles = [cl + base for cl in classes]

        if one_class is None:

            for i in range(len(df)):
                self._plot_shap_summary(df[i], top_n, titles[i], xlabel, show)

            return

        # If just one class passed, treat as just one class
        else:
            df = df[one_class]

            if title == 'default':
                title = titles[one_class]

    self._plot_shap_summary(df, top_n, title, xlabel, show)


def _plot_shap_summary(self, shap_df, top_n, title, xlabel, show):

    try:
        import shap
    except ImportError:
        raise ImportError('You must have shap installed to use shap')

    shap_cols, shap_inds = list(shap_df), shap_df.index
    data = self.all_data.loc[shap_inds][shap_cols]

    shap.summary_plot(np.array(shap_df), data, max_display=top_n, show=False)

    if title == 'default':
        title = 'Shap Summary Top ' + str(top_n) + ' Features'

    plt.title(title)

    if xlabel != 'default':
        plt.xlabel(xlabel)

    self._plot(title, show)



"""
ML_Helpers.py
====================================
File with various ML helper functions for BPt.
These are non-class functions that are used in _ML.py and Scoring.py
"""
import numpy as np
import inspect
from .Default_Params import get_base_params, proc_params
from copy import deepcopy
import nevergrad as ng
from ..main.Input_Tools import is_special, Select
from nevergrad.parametrization.core import Constant


def compute_micro_macro(scores, n_repeats, n_splits, weights=None):
    '''Compute and return scores, as computed froma repeated k-fold.

    Parameters
    ----------
    scores : list or array-like
        Should contain all of the scores
        and have a length of `n_repeats` * `n_splits`

    n_repeats : int
        The number of repeats

    n_splits : int
        The number of splits per repeat

    Returns
    ----------
    float
        The mean macro score

    float
        The standard deviation of the micro score

    float
        The standard deviation of the macro score
    '''

    r_scores = np.reshape(np.array(scores), (n_repeats, n_splits))

    if weights is None:
        macro_scores = np.mean(r_scores, axis=1)
    else:
        r_weights = np.reshape(np.array(weights), (n_repeats, n_splits))
        macro_scores = np.average(r_scores, weights=r_weights, axis=1)

    return (np.mean(macro_scores), np.std(scores), np.std(macro_scores))


def is_array_like(in_val):

    if hasattr(in_val, '__len__') and (not isinstance(in_val, str)) and \
     (not isinstance(in_val, dict)) and (not hasattr(in_val, 'fit')) and \
     (not hasattr(in_val, 'transform')):
        return True
    else:
        return False


def conv_to_list(in_val, amt=1):

    if in_val is None:
        return None

    if not is_array_like(in_val) or is_special(in_val):
        in_val = [in_val for i in range(amt)]

    return in_val


def proc_input(in_vals):
    '''Performs common preproc on a list of str's or
    a single str.'''

    if isinstance(in_vals, list):
        for i in range(len(in_vals)):
            in_vals[i] = proc_str_input(in_vals[i])
    else:
        in_vals = proc_str_input(in_vals)

    return in_vals


def proc_str_input(in_str):

    if not isinstance(in_str, str):
        return in_str

    # Make sure lower-case
    in_str = in_str.lower()

    # Remove regressor or classifier
    chunk_replace_dict = {' regressor': '',
                          ' classifier': ''}
    for chunk in chunk_replace_dict:
        in_str = in_str.replace(chunk, chunk_replace_dict[chunk])

    return in_str


def user_passed_param_check(params, obj_str, search_type):

    if isinstance(params, dict):
        if search_type is None:
            return deepcopy(params), {}
        else:
            return {}, deepcopy(proc_params(params, prepend=obj_str))
    return {}, {}


def proc_extra_params(obj, extra_params, non_search_params, params=None):

    if extra_params is None or len(extra_params) == 0:
        return non_search_params, params

    for key in extra_params:
        non_search_params[key] = deepcopy(extra_params[key])

        # Override value w/ extra params if also in params
        if params is not None and key in params:
            del params[key]

    return non_search_params, params


def get_obj_and_params(obj_str, OBJS, extra_params, params):

    # First get the object, and process the base params!
    try:
        obj, param_names = OBJS[obj_str]
    except KeyError:
        raise KeyError(repr(obj_str) + ' does not exist!')

    # If params is a str, change it to the relevant index
    if isinstance(params, str):
        try:
            params = param_names.index(params)
        except ValueError:
            print('str', params, 'passed, but not found as an option for',
                  obj_str)
            print('Setting to default base params setting instead!')
            params = 0

    # If passed param ind is a dict, assume that user passed
    if isinstance(params, dict):
        base_params = params.copy()

    # If not a dict passed, grab the param name, then params
    else:

        # Get the actual params
        try:
            param_name = param_names[params]
        except IndexError:
            print('Invalid param ind', params, 'passed for', obj_str)
            print('There are only', len(param_names), 'valid param options.')
            print('Setting to default base params setting instead!')
            param_name = param_names[0]

        base_params = get_base_params(param_name)

    # Process rest of params by search type, and w.r.t to extra params
    non_search_params, params =\
        process_params_by_type(obj, obj_str, base_params, extra_params)

    return obj, non_search_params, params


def process_params_by_type(obj, obj_str, base_params, extra_params):
    '''base params is either a dict or 0'''

    non_search_params, params = {}, {}

    # Start with params as copy of passed base_params
    params = deepcopy(base_params)

    # If params is passed as 0, treat as no search params
    if params == 0:
        params = {}

    # Extract any non-search params from params
    param_keys = list(params)
    for p in param_keys:

        try:
            module = params[p].__module__

            # If has a module, but no nevergrad in path, then not a
            # nevergrad param
            if 'nevergrad' not in module:
                non_search_params[p] = params.pop(p)

        # If no module, then not a nevergrad param
        except AttributeError:
            non_search_params[p] = params.pop(p)

    # Append obj_str to all params still in params
    # as these are the search params
    params = proc_params(params, prepend=obj_str)

    # Process extra params
    non_search_params, params =\
        proc_extra_params(obj,
                          extra_params=extra_params,
                          non_search_params=non_search_params,
                          params=params)

    return non_search_params, params


def get_possible_init_params(model):
    '''Helper function to grab the names of valid arguments to
    classes init

    Parameters
    ----------
    model : object
        The object to inspect

    Returns
    ----------
        All valid parameters to the model
    '''

    try:
        return model._get_param_names()
    except AttributeError:
        pos_params = dict(inspect.getmembers(model.__init__.__code__))
        return pos_params['co_varnames']


def get_possible_fit_params(model):
    '''Helper function to grab the names of valid arguments to
    classes fit method

    Parameters
    ----------
    model : object w/ fit method
        The model object to inspect

    Returns
    ----------
        All valid parameters to the model
    '''
    pos_params = dict(inspect.getmembers(model.fit.__code__))
    return pos_params['co_varnames']


def get_avaliable_by_type(AVALIABLE):

    avaliable_by_type = {}

    for pt in AVALIABLE:

        avaliable_by_type[pt] = set()
        for select in AVALIABLE[pt]:
            avaliable_by_type[pt].add(AVALIABLE[pt][select])

        avaliable_by_type[pt] = list(avaliable_by_type[pt])
        avaliable_by_type[pt].sort()

    return avaliable_by_type


def get_objects_by_type(problem_type, AVALIABLE=None, OBJS=None):

    avaliable_by_type = get_avaliable_by_type(AVALIABLE)

    objs = []
    for obj_str in avaliable_by_type[problem_type]:

        if 'basic ensemble' not in obj_str:
            obj = OBJS[obj_str][0]
            obj_params = OBJS[obj_str][1]
            objs.append((obj_str, obj, obj_params))

    return objs


def get_objects(OBJS):

    objs = []
    for obj_str in OBJS:

        obj = OBJS[obj_str][0]
        obj_params = OBJS[obj_str][1]
        objs.append((obj_str, obj, obj_params))

    return objs


def proc_problem_type(problem_type, avaliable_by_type):

    if problem_type is not None:
        problem_types = [problem_type]

    else:
        problem_types = list(avaliable_by_type)

    return problem_types


def f_array(in_array):
    return np.array(in_array).astype(float)


def find_ind(X, base_X_mask, X_r, r_ind, mask=True):

    r_dtype = X_r.dtype
    o_dtype = X.dtype

    if r_dtype != o_dtype and mask:
        ind = np.where(np.all(X[:, base_X_mask].astype(r_dtype) == X_r[r_ind],
                       axis=1))

    elif r_dtype != o_dtype:
        ind = np.where(np.all(X.astype(r_dtype) == X_r[r_ind], axis=1))

    elif mask:
        ind = np.where(np.all(X[:, base_X_mask] == X_r[r_ind], axis=1))

    else:
        ind = np.where(np.all(X == X_r[r_ind], axis=1))

    try:
        return ind[0][0]
    except IndexError:
        return None


def replace_with_in_params(params, original, replace):

    new_params = {}

    for key in params:
        new_params[key.replace(original, replace)] = params[key]

    return new_params


def type_check(ud):
    '''Check if a nevergrad dist'''

    def_dist = [ng.p.Log, ng.p.Scalar, ng.p.Choice, ng.p.TransitionChoice]
    for dd in def_dist:
        if isinstance(ud, dd):
            return True

    types_to_check = [int, float, list, tuple, str, bool, dict, set, Constant]

    for ttc in types_to_check:
        if isinstance(ud, ttc):
            return False

    return True


def proc_mapping(indx, mapping):

    if len(mapping) > 0 and len(indx) > 0:

        # If should proc list...
        if is_array_like(indx[0]):
            return [proc_mapping(i, mapping) for i in indx]

        else:
            new_indx = set()

            for i in indx:
                new = mapping[i]

                if new is None:
                    pass

                # If mapping points to a list of values
                elif isinstance(new, list):
                    for n in new:
                        if n is not None:
                            new_indx.add(n)
                else:
                    new_indx.add(new)

            # Sort, then return
            new_indx = sorted(list(new_indx))
            return new_indx

    else:
        return indx


def update_mapping(mapping, new_mapping):

    # Go through the mapping and update each key with the new mapping
    for key in mapping:

        val = mapping[key]

        if isinstance(val, list):

            new_vals = []
            for v in val:

                if v in new_mapping:

                    new_val = new_mapping[v]
                    if isinstance(new_val, list):
                        new_vals += new_val
                    else:
                        new_vals.append(new_val)

                else:
                    new_vals.append(v)

            as_set = set(new_vals)

            try:
                as_set.remove(None)
            except KeyError:
                pass

            mapping[key] = sorted(list(as_set))

        # Assume int if not list
        else:

            if val in new_mapping:
                mapping[key] = new_mapping[val]


def check_for_duplicate_names(objs_and_params):
    '''Checks for duplicate names within an objs_and_params type obj'''

    names = [c[0] for c in objs_and_params]

    # If any repeats
    if len(names) != len(set(names)):
        new_objs_and_params = []

        for obj in objs_and_params:
            name = obj[0]

            if name in names:

                cnt = 0
                used = [c[0] for c in new_objs_and_params]
                while name + str(cnt) in used:
                    cnt += 1

                # Need to change name within params also
                base_obj = obj[1][0]
                base_obj_params = obj[1][1]

                new_obj_params = {}
                for param_name in base_obj_params:

                    p_split = param_name.split('__')
                    new_param_name = p_split[0] + str(cnt)
                    new_param_name += '__' + '__'.join(p_split[1:])

                    new_obj_params[new_param_name] =\
                        base_obj_params[param_name]

                new_objs_and_params.append((name + str(cnt),
                                           (base_obj, new_obj_params)))

            else:
                new_objs_and_params.append(obj)

        return new_objs_and_params
    return objs_and_params


def proc_type_dep_str(in_strs, avaliable, problem_type):
    '''Helper function to perform str correction on
    underlying proble type dependent input, e.g., for
    ensemble_types, and to update extra params
    and check to make sure input is valid ect...'''

    as_arr = True
    if not is_array_like(in_strs):
        as_arr = False
        in_strs = [in_strs]

    if not check_avaliable(in_strs, avaliable, problem_type):
        in_strs = proc_input(in_strs)

        if not check_avaliable(in_strs, avaliable, problem_type):
            raise RuntimeError(in_strs, 'are not avaliable for '
                               'this problem type.'
                               'This may be due to the requested object '
                               'being an optional dependency! Check to make '
                               'sure you have the relevant python library '
                               'installed '
                               'and that the passed str contains no typos!')

    avaliable_by_type = get_a_by_type(avaliable, in_strs, problem_type)
    final_strs = [avaliable_by_type[in_str] for in_str in in_strs]

    if as_arr:
        return final_strs
    return final_strs[0]


def check_avaliable(in_strs, avaliable, problem_type):

    avaliable_by_type = get_a_by_type(avaliable, in_strs, problem_type)

    check = np.array([m in avaliable_by_type for
                      m in in_strs]).all()

    return check


def get_a_by_type(avaliable, in_strs, problem_type):

    avaliable_by_type = avaliable[problem_type]

    for s in in_strs:
        if 'Custom ' in s:
            avaliable_by_type[s] = s

    return avaliable_by_type


def param_len_check(names, params, _print=print):

    if isinstance(params, dict) and len(names) == 1:
        return params

    try:

        if len(params) > len(names):
            _print('Warning! More params passed than objs')
            _print('Extra params have been truncated.')
            return params[:len(names)]

    # If non list params here
    except TypeError:
        return [0 for i in range(len(names))]

    while len(names) != len(params):
        params.append(0)

    return params


def replace_model_name(base_estimator_params):

    new = {}

    for key in base_estimator_params:
        value = base_estimator_params[key]

        split_key = key.split('__')
        split_key[0] = 'estimator'

        new_key = '__'.join(split_key)
        new[new_key] = value

    return new


def get_avaliable_run_name(name, model_pipeline):

    if hasattr(model_pipeline, 'model'):
        model = model_pipeline.model
    else:
        model = model_pipeline

    if name is None or name == 'default':

        if isinstance(model, Select):
            name = 'select'
        elif isinstance(model, list):
            name = 'special'
        elif hasattr(model, 'obj'):
            if isinstance(model.obj, str):
                name = model.obj
            else:
                name = 'Custom'
        else:
            name = 'Custom'

    return name


def get_reverse_mapping(mapping):

    reverse_mapping = {}
    for m in mapping:
        key = mapping[m]

        if isinstance(key, list):
            for k in key:
                reverse_mapping[k] = m
        else:
            reverse_mapping[key] = m

    return reverse_mapping


def set_n_jobs(obj, n_jobs):

    # Call recursively for list
    if isinstance(obj, list) or isinstance(obj, tuple):
        for o in obj:
            set_n_jobs(o, n_jobs)

    # Check and set for n_jobs
    if hasattr(obj, 'n_jobs'):
        setattr(obj, 'n_jobs', n_jobs)

    # Also check for wrapper_n_jobs
    if hasattr(obj, 'wrapper_n_jobs'):
        setattr(obj, 'wrapper_n_jobs', n_jobs)



from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np
import warnings
import networkx as nx
from scipy.linalg import lstsq


class Identity(BaseEstimator, TransformerMixin):

    def __init__(self):
        '''This loader will simply flatten the input object, if not already.
        This loader is used to for example pass along loaded surfaces
        or volumes to a PCA or other simmilar transformation.'''
        pass

    def fit(self, X, y=None):
        pass

    def fit_transform(self, X, y=None):

        return self.transform(X)

    def transform(self, X):

        return X.flatten()


def load_surf(surf):
    '''Helper function to load a surface within BPt, w/ appropriate
    checks for important'''

    # If str, assume file path
    if isinstance(surf, str):

        # Only try to load with numpy if str ends with numpy ext
        if surf.endswith('.npy') or surf.endswith('.npz'):
            try:
                surf = np.load(surf)
                return surf
            except ValueError:
                pass

        # If not numpy, or numpy failed, try nilearn load surf data
        try:
            from nilearn.surface import load_surf_data
        except ImportError:
            raise ImportError('nilearn does not appear to be installed! ' +
                              'Install with "pip install nilearn", to load ' +
                              'surfaces from a file path.')

        surf = load_surf_data(surf)
        return surf

    # Keep as None for None
    elif surf is None:
        return None

    # If a Parc object
    elif hasattr(surf, 'get_parc'):
        return surf.get_parc(copy=True)

    # Otherwise assume either valid array-like passed
    else:
        return np.array(surf).copy()


class SurfLabels(BaseEstimator, TransformerMixin):

    def __init__(self, labels,
                 background_label=0,
                 mask=None,
                 strategy='mean',
                 vectorize=True):
        '''This class functions simmilar to NiftiLabelsMasker from nilearn,
        but instead is for surfaces (though it could work on a cifti
        image too).

        Parameters
        ----------
        labels : str or array-like
            This should represent an array, of the same size as the data
            dimension, as a mask
            with unique integer values for each ROI. You can also pass a str
            location in which
            to load in this array (though the saved file must be loadable by
            either numpy.load, or
            if not a numpy array, will try and load with
            nilearn.surface.load_surf_data(), which you
            will need nilearn installed to use.)

        background_labels : int, array-like of int, optional
            This parameter determines which label, if any,
            in the corresponding
            passed labels, should be treated as 'background'
            and therefore no ROI
            calculated for that value or values.
            You may pass either a single interger
            value, an array-like of integer values.

            If not background label is desired, just pass
            a label which doesn't exist in any of the data,
            e.g., -100.

            ::

                default = 0

        mask : None, str or array-like, optional
            This parameter allows you to optional pass a mask of values in
            which to not calculate ROI values for.
            This can be passed as a str or
            array-like of values (just like labels),
            and should be comprised of
            a boolean array (or 1's and 0's),
            where a value of 1 means that value
            will be ignored (set to background label)
            should be kept, and a value of 0,
            for that value should be masked away.
            This array should have the same length as the passed `labels`.

            ::

                default = None

        strategy: specific str, custom_func, optional
            This parameter dictates the function to be applied
            to each data's ROI's
            individually, e.g., mean to calculate the mean by ROI.

            If a str is passed, it must correspond to one of the below preset
            options:

            - 'mean'
                Calculate the mean with np.mean

            - 'sum'
                Calculate the sum with np.sum

            - 'min' or 'minimum
                Calculate the min value with np.min

            - 'max' or 'maximum
                Calculate the max value with np.max

            - 'std' or 'standard_deviation'
                Calculate the standard deviation with np.std

            - 'var' or  'variance'
                Calculate the variance with np.var

            If a custom function is passed, it must accept two arguments,
            custom_func(X_i, axis=data_dim), X_i, where X_i is a subjects data
            array where that subjects data corresponds to
            labels == some class i, and can potentially
            be either a 1D array or 2D array, and an axis argument
            to specify which axis is
            the data dimension (e.g., if calculating for a time-series
            [n_timepoints, data_dim], then data_dim = 1,
            if calculating for say stacked contrasts where
            [data_dim, n_contrasts], data_dim = 0, and lastly for a 1D
            array, data_dim is also 0.

            ::

                default = 'mean'

        vectorize : bool, optional
            If the returned array should be flattened to 1D. E.g., if the
            last step in a set of loader steps this should be True, if before
            a different step it may make sense to set to False.

            ::

                default = True
        '''

        self.labels = labels
        self.background_label = background_label
        self.mask = mask
        self.strategy = strategy
        self.vectorize = vectorize

    def fit(self, X, y=None):

        # Load mask if any
        self.mask_ = load_surf(self.mask)

        # Load labels
        self.labels_ = load_surf(self.labels)

        # Mask labels if mask
        if self.mask_ is not None:

            # Raise error if wrong shapes
            if len(self.mask_) != len(self.labels_):
                raise RuntimeError('length of mask must have '
                                   'the same length / shape as '
                                   'the length of labels!')
            self.labels_[self.mask_.astype(bool)] = self.background_label

        # X can either be a 1D surface, or a 2D surface
        # (e.g. - for timeseries or stacked contrasts)
        if len(X.shape) > 2:
            raise RuntimeError('X can be at most 2D.')

        if len(self.labels_) not in X.shape:
            raise RuntimeError('Size of labels not found in X. '
                               'Make sure your data is in the same '
                               'space as the labels you are using!')

        # Proc self.background_label, if int turn to np array
        if isinstance(self.background_label, int):
            self.background_label_ = np.array([self.background_label])

        # Otherwise, if already list-like, just cast to np array
        else:
            self.background_label_ = np.array(self.background_label)

        # Set the _non_bkg_unique as the valid labels to get ROIs for
        self.non_bkg_unique_ = np.setdiff1d(np.unique(self.labels_),
                                            self.background_label_)

        # Proc strategy if need be
        strats = {'mean': np.mean,
                  'median': np.median,
                  'sum': np.sum,
                  'minimum': np.min,
                  'min': np.min,
                  'maximum': np.max,
                  'max': np.max,
                  'standard_deviation': np.std,
                  'std': np.std,
                  'variance': np.var,
                  'var': np.var}

        if self.strategy in strats:
            self.strategy_ = strats[self.strategy]
        else:
            self.strategy_ = self.strategy

        return self

    def fit_transform(self, X, y=None):
        return self.fit(X, y).transform(X)

    def _check_fitted(self):
        if not hasattr(self, "labels_"):
            raise ValueError('It seems that SurfLabels has not been fitted. '
                             'You must call fit() before calling transform()')

    def transform(self, X):
        ''' If X has the both the same dimensions, raise warning'''

        self._check_fitted()

        if len(X.shape) == 2 and (X.shape[0] == X.shape[1]):
            warnings.warn('X was passed with the same length',
                          ' in each dimension, ',
                          'Assuming that axis=0 is the data dimension',
                          ' w/ vertex values')

        # The data dimension is just the dimension with
        # the same len as the label
        self.data_dim_ = X.shape.index(len(self.labels_))
        self.X_shape_ = X.shape

        # Get the ROI value for each label
        X_trans = []
        for i in self.non_bkg_unique_:

            if self.data_dim_ == 0:
                X_i = X[self.labels_ == i]
            else:
                X_i = X[:, self.labels_ == i]

            X_trans.append(self.strategy_(X_i, axis=self.data_dim_))

        if self.data_dim_ == 1:
            X_trans = np.stack(X_trans, axis=1)
        else:
            X_trans = np.array(X_trans)

        # Return based on vectorizes
        if not self.vectorize:
            return X_trans

        # Save original shape if vectorize called,
        # used for reverse transform
        self.original_shape_ = X_trans.shape
        return X_trans.flatten()

    def inverse_transform(self, X_trans):

        # Reverse the vectorize
        if self.vectorize:
            X_trans = X_trans.reshape(self.original_shape_)

        X = np.zeros(self.X_shape_, dtype=X_trans.dtype, order='C')

        if self.data_dim_ == 1:
            X_trans = np.rollaxis(X_trans, -1)
            X = np.rollaxis(X, -1)

        for i, label in enumerate(self.non_bkg_unique_):
            X[self.labels_ == label] = X_trans[i]

        if self.data_dim_ == 1:
            X = np.rollaxis(X, -1)

        return X


class SurfMaps(BaseEstimator, TransformerMixin):

    def __init__(self, maps, strategy='auto', mask=None, vectorize=True):
        '''Simmilar to NiftiMapsMasker from nilearn
        but for surfaces, and designed to work with BPt Loader.

        This object calculates the signal for each of the
        passed maps as extracted from the input during fit,
        and returns for each map a value.

        maps : str or array-like, optional
            This parameter represents the maps in which
            to apply to each surface, where the shape of
            the passed maps should be (# of vertex, # of maps)
            or in other words, the size of the data array in the first
            dimension and the number of maps
            (i.e., the number of outputted ROIs from fit)
            as the second dimension.

            You may pass maps as either an array-like,
            or the str file location of a numpy or other
            valid surface file format array in which to load.

        strategy : {'auto', 'ls', 'average'}, optional
            The stratgey in which the maps are used to extract
            signal. If 'ls' is selected, which stands for
            least squares, the least-squares solution will
            be used for each region.

            Alternatively, if 'average' is passed, then
            the weighted average value for each map
            will be computed.

            By default 'auto' will be selected,
            which will use 'average' if the passed
            maps contain only positive weights, and
            'ls' in the case that there are
            any negative values in the passed maps.

            Otherwise, you can set a specific strategy.
            In deciding which method to use,
            consider an example. Let's say the
            fit data X, and maps are

            ::

                data = np.array([1, 1, 5, 5])
                maps = np.array([[0, 0],
                                 [0, 0],
                                 [1, -1],
                                 [1, -1]])

            In this case, the 'ls' method would
            yield region signals [2.5, -2.5], whereas
            the weighted 'average' method, would yield
            [5, 5], notably ignoring the negative weights.
            This highlights an important limitation to the
            weighted averaged method, as it does not
            handle negative values well.

            On the other hand, consider changing the maps
            weights to

            ::

                data = np.array([1, 1, 5, 5])
                maps = np.array([[0, 1],
                                 [0, 2],
                                 [1, 0],
                                 [1, 0]])

                ls_sol = [5. , 0.6]
                average_sol = [5, 1]

            In this case, we can see that the weighted
            average gives a maybe more intuative summary
            of the regions. In general, it depends on
            what signal you are trying to summarize, and
            how you are trying to summarize it.

        mask : None, str or array-like, optional
            This parameter allows you to optional pass a mask of values in
            which to not calculate ROI values for.
            This can be passed as a str or
            array-like of values (just like maps),
            and should be comprised of
            a boolean array (or 1's and 0's),
            where a value of 1 means that value
            will be ignored (set to 0)
            should be kept, and a value of 0,
            for that value should be masked away.
            This array should have the same length as the passed `maps`.
            Specifically, where the shape of maps is (size, n_maps),
            the shape of mask should be (size).

            ::

                default = None

        vectorize : bool, optional
            If the returned array should be flattened to 1D. E.g., if this is
            the last step in a set of loader steps this should be True.
            Also note, if the surface data it is being applied to is 1D,
            then the output will be 1D regardless of this parameter.

            ::

                default = True
        '''

        self.maps = maps
        self.strategy = strategy
        self.mask = mask
        self.vectorize = vectorize

    def fit(self, X, y=None):

        # Load mask if any
        self.mask_ = load_surf(self.mask)

        # Load maps
        self.maps_ = load_surf(self.maps)

        # Make the maps if passed mask set to 0 in those spots
        if self.mask_ is not None:

            # Raise error if wrong shapes
            if len(self.mask_) != self.maps_.shape[0]:
                raise RuntimeError('length of mask must have '
                                   'the same length / shape as '
                                   'the first dimension of passed '
                                   'maps!')
            self.maps_[self.mask_.astype(bool)] = 0

        # X can either be a 1D surface, or a 2D surface
        # (e.g. - for timeseries or stacked contrasts)
        if len(X.shape) > 2:
            raise RuntimeError('X can be at most 2D.')

        # Check to make sure dimension of data is correct
        if self.maps_.shape[0] not in X.shape:
            raise RuntimeError('Size of labels not found in X. '
                               'Make sure your data is in the same '
                               'space as the labels you are using!')

        # Make sure strategy exists
        if self.strategy not in ['auto', 'ls', 'average']:
            raise RuntimeError('strategy must be '
                               '"ls", "average" or "auto"!')

        return self

    def fit_transform(self, X, y=None):
        return self.fit(X, y).transform(X)

    def _check_fitted(self):
        if not hasattr(self, "maps_"):
            raise ValueError('It seems that SurfMaps has not been fitted. '
                             'You must call fit() before calling transform()')

    def transform(self, X):

        self._check_fitted()

        if len(X.shape) == 2 and (X.shape[0] == X.shape[1]):
            warnings.warn('X was passed with the same length',
                          ' in each dimension, ',
                          'Assuming that axis=0 is the data dimension',
                          ' w/ vertex values')

        # The data dimension is just the dimension with
        # the first dimension of maps
        self.data_dim_ = X.shape.index(self.maps_.shape[0])
        self.X_shape_ = X.shape

        # If data in second dimension, transpose
        if self.data_dim_ == 1:
            X = X.T

        # Set strategy if auto
        strategy = self.strategy
        if strategy == 'auto':
            strategy = 'ls'
            if np.all(self.maps_ >= 0):
                strategy = 'average'

        # Run the correct transform based on strategy
        if strategy == 'ls':
            X_trans = self._transform_ls(X)
        elif strategy == 'average':
            X_trans = self._transform_average(X)
        else:
            X_trans = None

        # Always return as shape of extra data if any by
        # number of maps, with number of maps as last dimension

        # Return based on vectorize
        if not self.vectorize:
            return X_trans

        # Save original transformed output shape if vectorize
        self.original_shape_ = X_trans.shape
        return X_trans.flatten()

    def _transform_ls(self, X):
        '''X should be data points, or data points x stacked.'''

        X_trans = lstsq(self.maps_, X)[0]
        return X_trans.T

    def _transform_average(self, X):
        '''X should be data points, or data points x stacked.'''

        X_trans = []

        # For each map - take weighted average
        for m in range(self.maps_.shape[1]):

            try:
                X_trans.append(np.average(X, axis=0, weights=self.maps_[:, m]))
            except ZeroDivisionError:
                pass

        return np.array(X_trans).T

    def inverse_transform(self, X_trans):

        # Reverse the vectorize, if needed
        if self.vectorize:
            X_trans = X_trans.reshape(self.original_shape_)

        # For ls should be something like
        # np.dot(region_signals, maps.T)

        # For inverse transform weighted average,
        # make sure to include check for sum(weights) == 0
        # and to fill in that spot with zeros

        raise RuntimeError('Not Implemented Yet')


# Create wrapper for nilearn connectivity measure to make it
# work with 1 subject
try:
    from nilearn.connectome import ConnectivityMeasure

    class Connectivity(ConnectivityMeasure):

        def proc_X(self, X):

            if not isinstance(X, list):
                if len(np.shape(X)) == 2:
                    return [X]

            return X

        def fit(self, X, y=None):
            return super().fit(self.proc_X(X), y)

        def fit_transform(self, X, y=None):
            return super().fit_transform(self.proc_X(X), y)

        def transform(self, X):
            return super().transform(self.proc_X(X))

except ImportError:
    pass


class Networks(BaseEstimator, TransformerMixin):
    def __init__(self, threshold=.2, threshold_method='abs',
                 to_compute='avg_degree'):

        self.threshold = threshold
        self.threshold_method = threshold_method
        self.to_compute = to_compute

    def fit(self, X, y=None):
        '''X is a 2d correlation matrix'''

        if isinstance(self.to_compute, str):
            self.to_compute = [self.to_compute]

        try:
            import networkx
        except ImportError:
            raise ImportError(
                'To use this class, make sure you have networkx installed!')
        return self

    def fit_transform(self, X, y=None):
        return self.fit(X, y).transform(X)

    def _apply_threshold(self, X):

        if self.threshold_method == 'abs':
            return np.where(np.abs(X) >= self.threshold, 1, 0)
        elif self.threshold_method == 'pos':
            return np.where(X[0] >= self.threshold, 1, 0)
        elif self.threshold_method == 'neg':
            return np.where(X[0] <= self.threshold, 1, 0)
        elif self.threshold_method == 'density':
            top_n = round((len(np.triu(X).flatten())-len(X))/2*self.threshold)
            thres = sorted(np.triu(X).flatten(), reverse=True)[top_n]
            return np.where(X >= thres, 1, 0)

    def _threshold_check(self, X):

        while np.sum(self._apply_threshold(X)) == 0:
            print('warning setting threshold lower', self.threshold)
            self.threshold -= .01

    def transform(self, X):

        # Squeeze X
        X = np.squeeze(X)

        # Make sure the specified threshold doesn't break everything
        self._threshold_check(X)

        # Apply threshold
        X = self._apply_threshold(X)
        G = nx.from_numpy_matrix(X)

        func_dict = {'avg_cluster': nx.average_clustering,
                     'assortativity': nx.degree_assortativity_coefficient,
                     'global_eff': nx.global_efficiency,
                     'local_eff': nx.local_efficiency,
                     'transitivity': nx.transitivity,
                     'avg_eigenvector_centrality':
                     self._avg_eigenvector_centrality,
                     'avg_closeness_centrality':
                     self._avg_closeness_centrality,
                     'avg_degree': self._avg_degree,
                     'avg_triangles': self._avg_triangles,
                     'avg_pagerank': self._avg_pagerank,
                     'avg_betweenness_centrality':
                     self._avg_betweenness_centrality,
                     'avg_information_centrality':
                     self._avg_information_centrality,
                     'avg_shortest_path_length':
                     nx.average_shortest_path_length
                     }

        X_trans = []
        for compute in self.to_compute:
            X_trans += [func_dict[compute](G)]
        return np.array(X_trans)

    # Local
    def _avg_degree(self, G):
        avg_degree = np.mean([i[1] for i in nx.degree(G)])
        return avg_degree

    def _avg_triangles(self, G):
        avg_triangles = np.mean([nx.triangles(G)[i] for i in nx.triangles(G)])
        return avg_triangles

    def _avg_eigenvector_centrality(self, G):
        avg_eigenvector_centrality =\
            np.mean([nx.eigenvector_centrality_numpy(G)[
                     i] for i in nx.eigenvector_centrality_numpy(G)])
        return avg_eigenvector_centrality

    def _avg_closeness_centrality(self, G):
        avg_closeness_centrality = np.mean(
            [nx.closeness_centrality(G)[i] for i in nx.closeness_centrality(G)])
        return avg_closeness_centrality

    def _avg_betweenness_centrality(self, G):
        avg_betweenness_centrality = np.mean(
            [nx.betweenness_centrality(G)[i]
             for i in nx.betweenness_centrality(G)])
        return avg_betweenness_centrality

    def _avg_information_centrality(self, G):
        avg_information_centrality = np.mean(
            [nx.information_centrality(G)[i]
             for i in nx.information_centrality(G)])
        return avg_information_centrality

    def _avg_pagerank(self, G):
        avg_pagerank = np.mean([nx.pagerank(G)[i] for i in nx.pagerank(G)])
        return avg_pagerank



"""
Default_Params.py
====================================
File with different saved default parameter grids,
for various classifiers within BPt.
"""

from copy import deepcopy
from sklearn.feature_selection import (f_regression, f_classif,
                                       mutual_info_classif,
                                       mutual_info_regression, chi2)
import nevergrad as ng
import numpy as np


P = {}

cls_weight = "ng.p.TransitionChoice([None, 'balanced'])"
P['default'] = {}

# Models
P['base logistic'] =\
        {'max_iter': "1000",
         'multi_class': "'auto'",
         'penalty': "'none'",
         'class_weight': "None",
         'solver': "'lbfgs'"}

# Ridge classifier
P['base ridge'] = {'max_iter': "1000",
                   'penalty': "'l2'",
                   'solver': "'saga'"}

P['ridge C'] =\
        {'max_iter': "1000",
         'solver': "'saga'",
         'C': "ng.p.Log(lower=1e-5, upper=1e3)",
         'class_weight': cls_weight}

P['ridge C extra'] = P['ridge C'].copy()
P['ridge C extra']['max_iter'] =\
        "ng.p.Scalar(lower=1000, upper=10000).set_integer_casting()"
P['ridge C extra']['tol'] = "ng.p.Log(lower=1e-6, upper=.01)"

# Ridge regressor
P['base ridge regressor'] = {'max_iter': "1000",
                             'solver': "'lsqr'"}

P['ridge regressor dist'] = P['base ridge regressor'].copy()
P['ridge regressor dist']['alpha'] = "ng.p.Log(lower=1e-3, upper=1e5)"

# Lasso regressor
P['base lasso regressor'] = {'max_iter': "1000"}
P['lasso regressor dist'] = P['base lasso regressor'].copy()
P['lasso regressor dist']['alpha'] = "ng.p.Log(lower=1e-5, upper=1e5)"

# Lasso classifier
P['base lasso'] = P['base logistic'].copy()
P['base lasso']['solver'] = "'liblinear'"
P['base lasso']['penalty'] = "'l1'"

P['lasso C'] = P['base lasso'].copy()
P['lasso C']['C'] = "ng.p.Log(lower=1e-5, upper=1e3)"
P['lasso C']['class_weight'] = cls_weight

P['lasso C extra'] = P['lasso C'].copy()
P['lasso C extra']['max_iter'] =\
        "ng.p.Scalar(lower=1000, upper=10000).set_integer_casting()"
P['lasso C extra']['tol'] = "ng.p.Log(lower=1e-6, upper=.01)"

# Elastic net classifier
P['base elastic'] = P['base logistic'].copy()
P['base elastic']['penalty'] = "'elasticnet'"
P['base elastic']['l1_ratio'] = ".5"
P['base elastic']['solver'] = "'saga'"

P['elastic classifier'] = P['base elastic'].copy()
P['elastic classifier']['C'] = "ng.p.Log(lower=1e-5, upper=1e5)"
P['elastic classifier']['l1_ratio'] = "ng.p.Scalar(lower=.01, upper=1)"
P['elastic classifier']['class_weight'] = cls_weight

P['elastic clf v2'] = P['elastic classifier'].copy()
P['elastic clf v2']['C'] = "ng.p.Log(lower=1e-2, upper=1e5)"

P['elastic classifier extra'] = P['elastic classifier'].copy()
P['elastic classifier extra']['max_iter'] =\
        "ng.p.Scalar(lower=1000, upper=10000).set_integer_casting()"
P['elastic classifier extra']['tol'] = "ng.p.Log(lower=1e-6, upper=.01)"

# Elastic net regression
P['base elastic net'] = {'max_iter': "1000"}
P['elastic regression'] = P['base elastic net'].copy()
P['elastic regression']['alpha'] = "ng.p.Log(lower=1e-5, upper=1e5)"
P['elastic regression']['l1_ratio'] = "ng.p.Scalar(lower=.01, upper=1)"

P['elastic regression extra'] = P['elastic regression'].copy()
P['elastic regression extra']['max_iter'] =\
        "ng.p.Scalar(lower=1000, upper=10000).set_integer_casting()"
P['elastic regression extra']['tol'] = "ng.p.Log(lower=1e-6, upper=.01)"


P['base huber'] = {'epsilon': "1.35"}
P['base gnb'] = {'var_smoothing': "1e-9"}

P['base knn'] = {'n_neighbors': "5"}
P['knn dist'] = {'weights': "ng.p.TransitionChoice(['uniform', 'distance'])",
                 'n_neighbors':
                 "ng.p.Scalar(lower=2, upper=25).set_integer_casting()"}
P['base knn regression'] = P['base knn'].copy()
P['knn dist regression'] = P['knn dist'].copy()

P['base dt'] = {}
P['dt dist'] = {'max_depth':
                "ng.p.Scalar(lower=1, upper=30).set_integer_casting()",
                'min_samples_split':
                "ng.p.Scalar(lower=2, upper=50).set_integer_casting()"}

P['dt classifier dist'] = P['dt dist'].copy()
P['dt classifier dist']['class_weight'] = cls_weight

P['base linear'] = {'fit_intercept': "True"}

P['base rf'] = {'n_estimators': "100"}
P['base rf regressor'] = P['base rf'].copy()

n_estimators = "ng.p.Scalar(init=100, lower=3, upper=500).set_integer_casting()"
depths = "ng.p.TransitionChoice([None, ng.p.Scalar(init=25, lower=2, upper=200).set_integer_casting()])"

P['rf dist'] = {'n_estimators': n_estimators,
                'max_depth': depths,
                'max_features': "ng.p.Scalar(lower=.1, upper=1.0)",
                'min_samples_split': "ng.p.Scalar(lower=.1, upper=1.0)",
                'bootstrap': "True"}

P['rf classifier dist'] = P['rf dist'].copy()
P['rf classifier dist']['class_weight'] = cls_weight

# Light gbm params
P['base lgbm'] = {'silent': "True"}

reg = "ng.p.TransitionChoice([0, ng.p.Log(lower=1e-5, upper=1)])"

P['lgbm dist1'] =\
        {'silent': "True",
         'boosting_type': "ng.p.TransitionChoice(['gbdt', 'dart', 'goss'])",
         'n_estimators': n_estimators,
         'num_leaves':
         "ng.p.Scalar(init=20, lower=6, upper=80).set_integer_casting()",
         'min_child_samples':
         "ng.p.Scalar(lower=10, upper=500).set_integer_casting()",
         'min_child_weight':
         "ng.p.Log(lower=1e-5, upper=1e4)",
         'subsample':
         "ng.p.Scalar(lower=.3, upper=.95)",
         'colsample_bytree':
         "ng.p.Scalar(lower=.3, upper=.95)",
         'reg_alpha': reg,
         'reg_lambda': reg}

P['lgbm dist2'] =\
        {'silent': "True",
         'lambda_l2': "0.001",
         'boosting_type': "ng.p.TransitionChoice(['gbdt', 'dart'])",
         'min_child_samples':
         "ng.p.TransitionChoice([1, 5, 7, 10, 15, 20, 35, 50, 100, 200, 500, 1000])",
         'num_leaves':
         "ng.p.TransitionChoice([2, 4, 7, 10, 15, 20, 25, 30, 35, 40, 50, 65, 80, 100, 125, 150, 200, 250])",
         'colsample_bytree':
         "ng.p.TransitionChoice([0.7, 0.9, 1.0])",
         'subsample':
         "ng.p.Scalar(lower=.3, upper=1)",
         'learning_rate':
         "ng.p.TransitionChoice([0.01, 0.05, 0.1])",
         'n_estimators':
         "ng.p.TransitionChoice([5, 20, 35, 50, 75, 100, 150, 200, 350, 500, 750, 1000])"}

P['lgbm dist3'] = {'silent': "True",
                   'n_estimators': "1000",
                   'early_stopping_rounds': "150",
                   'eval_split': ".2",
                   'boosting_type': '"gbdt"',
                   'learning_rate':
                   'ng.p.Log(lower=5e-3, upper=.2, init=.1)',
                   'colsample_bytree':
                   'ng.p.Scalar(lower=.75, upper=1, init=1)',
                   'min_child_samples':
                   "ng.p.Scalar(lower=2, upper=30, init=20).set_integer_casting()",
                   'num_leaves':
                   "ng.p.Scalar(lower=16, upper=96, init=31).set_integer_casting()"}

P['lgbm classifier dist1'] = P['lgbm dist1'].copy()
P['lgbm classifier dist1']['class_weight'] = cls_weight

P['lgbm classifier dist2'] = P['lgbm dist2'].copy()
P['lgbm classifier dist2']['class_weight'] = cls_weight

P['lgbm classifier dist3'] = P['lgbm dist3'].copy()
P['lgbm classifier dist3']['class_weight'] = cls_weight

P['base xgb'] = {'verbosity': "0",
                 'objective': "'reg:squarederror'"}

P['base xgb classifier'] = P['base xgb'].copy()
P['base xgb classifier']['objective'] = "'binary:logistic'"

P['xgb dist1'] =\
        {'verbosity': "0",
         'objective': "'reg:squarederror'",
         'n_estimators': n_estimators,
         'min_child_weight': "ng.p.Log(lower=1e-5, upper=1e4)",
         'subsample': "ng.p.Scalar(lower=.3, upper=.95)",
         'colsample_bytree': "ng.p.Scalar(lower=.3, upper=.95)",
         'reg_alpha': reg,
         'reg_lambda': reg}

P['xgb dist2'] =\
        {'verbosity': "0",
         'objective': "'reg:squarederror'",
         'max_depth': depths,
         'learning_rate': "ng.p.Scalar(lower=.01, upper=.5)",
         'n_estimators': "ng.p.Scalar(lower=3, upper=500).set_integer_casting()",
         'min_child_weight': "ng.p.TransitionChoice([1, 5, 10, 50])",
         'subsample': "ng.p.Scalar(lower=.5, upper=1)",
         'colsample_bytree': "ng.p.Scalar(lower=.4, upper=.95)"}

P['xgb dist3'] =\
        {'verbosity': "0",
         'objective': "'reg:squarederror'",
         'learning_rare': "ng.p.Scalar(lower=.005, upper=.3)",
         'min_child_weight': "ng.p.Scalar(lower=.5, upper=10)",
         'max_depth': "ng.p.TransitionChoice(np.arange(3, 10))",
         'subsample': "ng.p.Scalar(lower=.5, upper=1)",
         'colsample_bytree': "ng.p.Scalar(lower=.5, upper=1)",
         'reg_alpha': "ng.p.Log(lower=.00001, upper=1)"}

P['xgb classifier dist1'] = P['xgb dist1'].copy()
P['xgb classifier dist1']['objective'] = "'binary:logistic'"

P['xgb classifier dist2'] = P['xgb dist2'].copy()
P['xgb classifier dist2']['objective'] = "'binary:logistic'"

P['xgb classifier dist3'] = P['xgb dist3'].copy()
P['xgb classifier dist3']['objective'] = "'binary:logistic'"

P['base gp regressor'] = {'n_restarts_optimizer': "5",
                          'normalize_y': "True"}
P['base gp classifier'] = {'n_restarts_optimizer': "5"}

# probability = True
P['base svm'] = {'kernel': "'rbf'",
                 'gamma': "'scale'"}

P['svm dist'] = P['base svm'].copy()
P['svm dist']['C'] = "ng.p.Log(lower=1e-4, upper=1e4)"
P['svm dist']['gamma'] = "ng.p.Log(lower=1e-6, upper=1)"

P['base svm classifier'] = P['base svm'].copy()
P['base svm classifier']['probability'] = "True"

P['svm classifier dist'] = P['svm dist'].copy()
P['svm classifier dist']['probability'] = "True"
P['svm classifier dist']['class_weight'] = cls_weight


# Define different choices for the mlp
P['base mlp'] = {}

batch_size =\
        "ng.p.TransitionChoice(['auto', ng.p.Scalar(init=200, lower=50," +\
        " upper=400).set_integer_casting()])"

P['mlp dist 1 layer'] =\
        {'hidden_layer_sizes':
         "ng.p.Scalar(init=100, lower=2, upper=300).set_integer_casting()",
         'activation':
         "ng.p.TransitionChoice(['identity', 'logistic', 'tanh', 'relu'])",
         'alpha': "ng.p.Log(lower=1e-5, upper=1e2)",
         'batch_size': batch_size,
         'learning_rate':
         "ng.p.TransitionChoice(['constant', 'invscaling', 'adaptive'])",
         'learning_rate_init': "ng.p.Log(lower=1e-5, upper=1e2)",
         'max_iter':
         "ng.p.Scalar(init=200, lower=100, upper=1000).set_integer_casting()",
         'beta_1': "ng.p.Scalar(init=.9, lower=.1, upper=.99)",
         'beta_2': "ng.p.Scalar(init=.999, lower=.1, upper=.9999)"}

P['mlp dist es 1 layer'] = P['mlp dist 1 layer'].copy()
P['mlp dist es 1 layer']['early_stopping'] = "True"
P['mlp dist es 1 layer']['n_iter_no_change'] =\
        "ng.p.Scalar(lower=5, upper=50)"

two_layer = "ng.p.Array(init=(100, 100)).set_mutation(sigma=50)" +\
            ".set_bounds(lower=1, upper=300).set_integer_casting()"

P['mlp dist 2 layer'] = P['mlp dist 1 layer'].copy()
P['mlp dist 2 layer']['hidden_layer_sizes'] = two_layer

P['mlp dist es 2 layer'] = P['mlp dist es 1 layer'].copy()
P['mlp dist 2 layer']['hidden_layer_sizes'] = two_layer

three_layer = "ng.p.Array(init=(100, 100, 100)).set_mutation(sigma=50)" +\
              ".set_bounds(lower=1, upper=300).set_integer_casting()"

P['mlp dist 3 layer'] = P['mlp dist 1 layer'].copy()
P['mlp dist 3 layer']['hidden_layer_sizes'] = three_layer

P['mlp dist es 3 layer'] = P['mlp dist es 1 layer'].copy()
P['mlp dist 3 layer']['hidden_layer_sizes'] = three_layer

P['base linear svc'] = {'max_iter': "1000"}

P['linear svc dist'] = P['base linear svc'].copy()
P['linear svc dist']['C'] = "ng.p.Log(lower=1e-4, upper=1e4)"
P['linear svc dist']['class_weight'] = cls_weight

P['base linear svr'] = {'loss': "'epsilon_insensitive'",
                        'max_iter': "1000"}

P['linear svr dist'] = P['base linear svr'].copy()
P['linear svr dist']['C'] = "ng.p.Log(lower=1e-4, upper=1e4)"

P['base sgd'] = {'loss': "'squared_loss'"}

loss_choice = "ng.p.TransitionChoice(['hinge', 'log', " +\
              "'modified_huber', 'squared_hinge', 'perceptron'])"

lr_choice = "ng.p.TransitionChoice(['optimal', 'invscaling', " +\
            "'adaptive', 'constant'])"

P['sgd classifier big search'] =\
        {'loss': loss_choice,
         'penalty': "ng.p.TransitionChoice(['l2', 'l1', 'elasticnet'])",
         'alpha': "ng.p.Log(lower=1e-5, upper=1e2)",
         'l1_ratio': "ng.p.Scalar(lower=.01, upper=1)",
         'max_iter': "1000",
         'learning_rate': lr_choice,
         'eta0': "ng.p.Log(lower=1e-6, upper=1e3)",
         'power_t': "ng.p.Scalar(lower=.1, upper=.9)",
         'early_stopping': "ng.p.TransitionChoice([False, True])",
         'validation_fraction': "ng.p.Scalar(lower=.05, upper=.5)",
         'n_iter_no_change': "ng.p.TransitionChoice(np.arange(2, 20))",
         'class_weight': cls_weight}

# Make elastic net version
P['sgd elastic'] =\
        {'loss': "'squared_epsilon_insensitive'",
         'penalty': "'elasticnet'",
         'alpha': "ng.p.Log(lower=1e-5, upper=1e5)",
         'l1_ratio': "ng.p.Scalar(lower=.01, upper=1)"}

P['sgd elastic classifier'] = P['sgd elastic'].copy()
P['sgd elastic classifier']['class_weight'] = cls_weight

# Auto gluon
P['pt binary'] = {'problem_type': "'binary'"}
P['pt multiclass'] = {'problem_type': "'multiclass'"}
P['pt regression'] = {'problem_type': "'regression'"}


# Transformers
P['pca var search'] =\
        {'n_components': "ng.p.Scalar(init=.75, lower=.1, upper=.99)",
         'svd_solver': "'full'"}

P['ohe'] =\
        {'sparse': "False",
         'handle_unknown': "'ignore'"}

# Scalers
P['base standard'] = {'with_mean': "True",
                      'with_std': "True"}

P['base minmax'] = {'feature_range': "(0, 1)"}

P['base robust'] = {'quantile_range': "(5, 95)"}

P['base winsorize'] = {'quantile_range': "(1, 99)"}

P['robust gs'] =\
        {'quantile_range':
         "ng.p.TransitionChoice([(x, 100-x) for x in np.arange(1, 40)])"}

P['winsorize gs'] =\
        {'quantile_range':
         "ng.p.TransitionChoice([(x, 100-x) for x in np.arange(1, 40)])"}

P['base yeo'] = {'method': "'yeo-johnson'",
                 'standardize': "True"}

P['base boxcox'] = {'method': "'box-cox'",
                    'standardize': "True"}

P['base quant norm'] = {'output_distribution': "'normal'"}

P['base quant uniform'] = {'output_distribution': "'uniform'"}

# Feat Selectors
P['base univar fs regression'] = {'score_func': "f_regression",
                                  'percentile': "50"}

P['univar fs regression dist'] = {'score_func': "f_regression",
                                  'percentile':
                                  "ng.p.Scalar(init=50, lower=1, upper=99)"}

P['univar fs regression dist2'] = {'score_func': "f_regression",
                                   'percentile':
                                   "ng.p.Scalar(init=75, lower=50, upper=99)"}


P['base univar fs classifier'] = {'score_func': "f_classif",
                                  'percentile': "50"}

P['univar fs classifier dist'] = {'score_func': "f_classif",
                                  'percentile':
                                  "ng.p.Scalar(init=50, lower=1, upper=99)"}

P['univar fs classifier dist2'] = {'score_func': "f_classif",
                                   'percentile':
                                   "ng.p.Scalar(init=75, lower=50, upper=99)"}


P['base rfe'] = {'n_features_to_select': "None"}

P['rfe num feats dist'] = {'n_features_to_select':
                           "ng.p.Scalar(init=.5, lower=.1, upper=.99)"}

P['random'] = {'mask': "'sets as random features'"}
P['searchable'] = {'mask': "'sets as hyperparameters'"}

# Imputers
P['mean imp'] = {'strategy': "'mean'"}
P['median imp'] = {'strategy': "'median'"}
P['most freq imp'] = {'strategy': "'most_frequent'"}
P['constant imp'] = {'strategy': "'constant'"}
P['iterative imp'] = {'initial_strategy': "'mean'",
                      'skip_complete': "True"}

# Samplers
P['base no change sampler'] = {'sampler_type': "'no change'"}

P['base special sampler'] = {'sampler_type': "'special'"}

P['base change sampler'] = {'sampler_type': "'change'"}


P['rus binary ratio'] = P['base no change sampler'].copy()
P['rus binary ratio']['sampling_strategy'] =\
        "ng.p.Scalar(init=.8, lower=.1, upper=1.2)"


# Ensemblers
P['stacking default'] = {'cv': "3"}

P['voting classifier'] = {'voting': "'soft'"}

# Feat Importances
P['base shap'] =\
        {'shap__global__avg_abs': "False",
         'shap__linear__feature_dependence': "'independent'",
         'shap__linear__nsamples': "1000",
         'shap__tree__feature_perturbation': "'tree_path_dependent'",
         'shap__tree__model_output': "'margin'",
         'shap__tree__tree_limit': "None",
         'shap__kernel__nkmean': "10",
         'shap__kernel__nsamples': "'auto'",
         'shap__kernel__l1_reg': "'aic'"}

P['base perm'] = {'perm__n_perm': "10"}

PARAMS = {}
for param in P:
    PARAMS[param] = {}
    for p in P[param]:
        try:
            PARAMS[param][p] = eval(P[param][p])
        except TypeError:
            PARAMS[param][p] = P[param][p]


def get_base_params(str_indicator):

    base_params = deepcopy(PARAMS[str_indicator])
    return base_params


def proc_params(base_params, prepend=None):

    if isinstance(base_params, int):
        return {}

    params = {prepend + '__' + key: base_params[key] for key in
              base_params}

    return params


def show(str_indicator):

    params = PARAMS[str_indicator].copy()

    if len(params) == 0:
        print('None')

    for key in params:
        print(key, ': ', sep='', end='')

        value = params[key]
        print(value)



import os
import tempfile
import numpy as np
from ...helpers.Data_File import Data_File
from ...helpers.Data_Scopes import Data_Scopes
from ...main.Params_Classes import Problem_Spec
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_selection._base import SelectorMixin


class ToFixedTransformer(BaseEstimator, TransformerMixin):

    def __init__(self, to):
        self.to = to

    def fit(self, X, y):
        return self

    def transform(self, X):
        X_trans = np.zeros(X.shape)
        X_trans[:] = self.to
        return X_trans


class FakeSelector(SelectorMixin, BaseEstimator):

    def __init__(self, mask):
        self.mask = mask

    def fit(self, X, y):
        return self

    def _get_support_mask(self):
        return self.mask


def get_temp_files(n):

    temp_dr = tempfile.gettempdir()
    return [os.path.join(temp_dr, 'test_' + str(i) + '.npy') for i in range(n)]


def get_fake_mapping(n):

    locs = get_temp_files(n)

    mapping = {}
    for i, loc in enumerate(locs):
        data = np.zeros((2, 2))
        data[:] = i

        np.save(loc, data)

        mapping[i] = Data_File(loc=loc, load_func=np.load)

    return mapping


def clean_fake_mapping(n):

    locs = get_temp_files(n)
    for loc in locs:
        os.unlink(loc)


def get_fake_data_scopes(data_keys=None,
                         cat_keys=None):

    if data_keys is None:
        data_keys = []

    if cat_keys is None:
        cat_keys = []

    data_scopes =\
        Data_Scopes(data_keys=data_keys,
                    data_file_keys=[],
                    cat_keys=cat_keys,
                    strat_keys=[],
                    covars_keys=[],
                    file_mapping=None)

    fake_ps = Problem_Spec(target='target', scope='all')
    data_scopes.set_all_keys(fake_ps)

    return data_scopes



from .ML_Helpers import conv_to_list


class Data_Scopes():

    def __init__(self, data_keys, data_file_keys,
                 cat_keys, strat_keys, covars_keys,
                 file_mapping=None):
        '''Class to hold and handle the different scope level /
         all keys type fields'''

        self.data_keys = data_keys
        self.data_file_keys = data_file_keys
        self.cat_keys = cat_keys
        self.strat_keys = strat_keys
        self.covars_keys = covars_keys

        if file_mapping is None:
            file_mapping = {}
        self.file_mapping = file_mapping

    def set_all_keys(self, spec):

        # Keys at end used to include strat, now have only be targt
        self.keys_at_end = conv_to_list(spec.target)

        # Set all keys to data keys + covars keys + target keys at end to start
        self.all_keys =\
            self.data_keys + self.covars_keys + self.keys_at_end

        # Filter all_keys by scope, and set new all keys
        scoped_keys = self.get_keys_from_scope(spec.scope)
        self.all_keys = scoped_keys + self.keys_at_end

        # Store total number of valid keys
        self.num_feat_keys = len(self.all_keys) - len(self.keys_at_end)

        return self.all_keys

    def get_keys_from_scope(self, scope):

        from ..helpers.VARS import SCOPES

        if isinstance(scope, str) and scope in SCOPES:

            if scope == 'float':
                keys = [k for k in self.all_keys if k not in self.cat_keys]

            elif scope == 'data':
                keys = self.data_keys.copy()

            elif scope == 'data files':
                keys = self.data_file_keys.copy()

            elif scope == 'float covars' or scope == 'fc':
                keys = [k for k in self.all_keys if
                        k not in self.cat_keys and
                        k not in self.data_keys]

            elif scope == 'all' or scope == 'n':
                keys = self.all_keys.copy()

            elif scope == 'cat' or scope == 'categorical':
                keys = self.cat_keys.copy()

            elif scope == 'covars':
                keys = self.covars_keys.copy()

            else:
                keys = None
                print('Should never reach here.')
                pass

        # If not a valid passed scope, should be list like
        else:

            scope = conv_to_list(scope)

            keys, restrict_keys = [], []
            for key in scope:

                # If a valid scope, add to keys
                if key in SCOPES:
                    keys += self.get_keys_from_scope(key)

                # If a passed column name, add to keys
                elif key in self.all_keys:
                    keys.append(key)

                # Otherwise append to restrict keys
                else:
                    restrict_keys.append(key)

            # Restrict all keys by the stub strs passed that arn't scopes
            #  or valid column names
            if len(restrict_keys) > 0:
                keys += [a for a in self.all_keys if
                         all([r in a for r in restrict_keys])]

            # Get rid of repeats if any
            keys = list(set(keys))

        # Need to remove all strat keys + target_keys, if there regardless
        for key in self.keys_at_end:

            try:
                keys.remove(key)
            except ValueError:
                pass

        # Return sorted keys (for reproducible results)
        return sorted(keys)

    def get_train_inds_from_keys(self, keys):
        '''Assume target is always last within self.all_keys ...
        maybe a bad assumption.'''

        inds = [self.all_keys.index(k) for k in keys if k in self.all_keys]
        return inds

    def get_inds_from_scope(self, scope):
        '''Return inds from scope'''

        # First get keys from scope
        keys = self.get_keys_from_scope(scope)

        # Then change to inds
        inds = self.get_train_inds_from_keys(keys)

        return inds



from sklearn.feature_selection import RFE
from sklearn.base import BaseEstimator
from sklearn.feature_selection._base import SelectorMixin
import numpy as np
from ..helpers.ML_Helpers import proc_mapping


class RFE_Wrapper(RFE):
    def fit(self, X, y):
        '''Override the fit function from base
           specifically allow passing in float % to keep.
        '''

        if isinstance(self.n_features_to_select, float):

            if self.n_features_to_select > 1:
                self.n_features_to_select = round(self.n_features_to_select)

            else:
                if self.n_features_to_select <= 0:
                    self.n_features_to_select = 1

                if self.n_features_to_select < 1:
                    divide_by = self.n_features_to_select ** -1
                    self.n_features_to_select = X.shape[1] // divide_by

        return self._fit(X, y)


class FeatureSelector(SelectorMixin, BaseEstimator):

    def __init__(self, mask='sets as random features'):
        ''' Custom BPt feature selector for
        integrating in feature selection with a hyper-parameter search.

        Parameters
        ----------
        mask : {'sets as random features', 'sets as hyperparameters'}
            - 'sets as random features': Use random features.

            - 'sets as hyperparameters': Each feature is set as a \
            hyperparameter, such that the parameter search can \
            tune if each feature is included or not.


        '''

        self.mask = mask
        self.name = 'selector'

    def _proc_mapping(self, mapping):

        if hasattr(self, '_mapping'):
            return
        else:
            self._mapping = mapping

        if len(mapping) > 0:

            mask_0 = np.where(~self.mask)[0]
            mask_1 = np.where(self.mask)[0]

            mask_0 = proc_mapping(mask_0, mapping)
            mask_1 = proc_mapping(mask_1, mapping)

            if len(mask_0) > 0:
                mx_0 = np.max(mask_0)
            else:
                mx_0 = 0

            if len(mask_1) > 0:
                mx_1 = np.max(mask_1)
            else:
                mx_1 = 0

            max_ind = np.max([mx_0, mx_1])

            self.mask = np.zeros(max_ind+1, dtype='bool')
            self.mask[mask_1] = True
        return

    def fit(self, X, y=None, mapping=None):

        if mapping is None:
            mapping = {}

        self.mask = np.array(self.mask)

        threshold = .5
        while np.sum(self.mask >= threshold) == 0:
            threshold -= .001

        self.mask = self.mask >= threshold
        self._proc_mapping(mapping)

        return self

    def _get_support_mask(self):
        return self.mask
